# Copyright(C) 2010-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
from pathlib import Path
import logging
import warnings

from typing import List, Callable, Dict, Iterator

from woob import __version__
from woob.capabilities.base import Capability
from woob.core.backendscfg import BackendsConfig
from woob.core.bcall import BackendsCall
from woob.core.modules import ModulesLoader, RepositoryModulesLoader
from woob.core.repositories import Repositories, IProgress, PrintProgress
from woob.core.requests import RequestsManager
from woob.core.scheduler import IScheduler, Scheduler
from woob.exceptions import ModuleLoadError
from woob.tools.backend import Module
from woob.tools.config.iconfig import ConfigError
from woob.tools.log import getLogger
from woob.tools.misc import classproperty
from woob.tools.storage import IStorage


__all__ = ['WoobBase', 'Woob']


class VersionsMismatchError(ConfigError):
    pass


class WoobBase:
    """
    Woob class to load modules from a specific path, without deal with woob
    remote repositories.

    It provides methods to build backends or call methods on all loaded
    backends.

    You should use this class when you want to build an application
    using Woob as a library, without using the standard modules nor
    the automatic module download and update machanism. When using
    WoobBase, you have to explicitely provide module paths and deal
    yourself with backend configuration.

    :param modules_path: path to directory containing modules.
    :type modules_path: :class:`str`
    :param storage: provide a storage where backends can save data
    :type storage: :class:`woob.tools.storage.IStorage`
    :param scheduler: what scheduler to use; default is :class:`woob.core.scheduler.Scheduler`
    :type scheduler: :class:`woob.core.scheduler.IScheduler`
    """

    @classproperty
    def VERSION(self):
        warnings.warn('Use woob.__version__ instead.', DeprecationWarning, stacklevel=2)
        return __version__

    def __init__(self,
                 modules_path: str | None = None,
                 storage: IStorage | None = None,
                 scheduler: IScheduler | None = None):
        self.logger = getLogger('woob')
        self.backend_instances: Dict[str, Module] = {}
        self.requests = RequestsManager()

        self.modules_path = modules_path
        self.modules_loader = self.build_modules_loader()

        if scheduler is None:
            scheduler = Scheduler()
        self.scheduler = scheduler

        self.storage = storage

    def __deinit__(self):
        self.deinit()

    def deinit(self):
        """
        Call this method when you stop using Woob, to
        properly unload all correctly.
        """
        self.unload_backends()

    def build_modules_loader(self) -> ModulesLoader:
        """
        Build the module loader for the current application.

        This can be overridden by children to avoid overriding an already
        existing modules loader.

        :rtype: ModulesLoader
        """
        return ModulesLoader(self.modules_path, __version__)

    def build_backend(self,
                      module_name: str,
                      params: Dict[str, str] | None = None,
                      storage: IStorage | None = None,
                      name: str | None = None,
                      nofail: bool = False,
                      logger: logging.Logger | None = None) -> Module:
        """
        Create a backend.

        It does not load it into the Woob object, so you are responsible for
        deinitialization and calls.

        :param module_name: name of module
        :param params: parameters to give to backend
        :type params: :class:`dict`
        :param storage: storage to use
        :type storage: :class:`woob.tools.storage.IStorage`
        :param name: name of backend
        :type name: :class:`str`
        :param nofail: if true, this call can't fail
        :type nofail: :class:`bool`
        :param logger: logger to use
        :type logger: :class:`logging.Logger`
        :rtype: :class:`woob.tools.backend.Module`
        """
        module = self.modules_loader.get_or_load_module(module_name)

        backend_instance = module.create_instance(self, name or module_name, params or {}, storage, nofail, logger=logger or self.logger)
        return backend_instance

    class LoadError(Exception):
        """
        Raised when a backend is unabled to load.

        :param backend_name: name of backend we can't load
        :param exception: exception object
        """

        def __init__(self, backend_name: str, message: str):
            super().__init__(message)
            self.backend_name = backend_name

    def load_backend(self,
                     module_name: str,
                     name: str,
                     params: dict | None = None,
                     storage: IStorage | None = None,
                     nofail: bool = False):
        """
        Load a backend.

        :param module_name: name of module to load
        :type module_name: :class:`str`:
        :param name: name of instance
        :type name: :class:`str`
        :param params: parameters to give to backend
        :type params: :class:`dict`
        :param storage: storage to use
        :type storage: :class:`woob.tools.storage.IStorage`
        :param nofail: if true, this call can't fail
        :type nofail: :class:`bool`
        :rtype: :class:`woob.tools.backend.Module`
        """
        if name is None:
            name = module_name

        if name in self.backend_instances:
            raise self.LoadError(name, 'A loaded backend already named "%s"' % name)

        backend = self.build_backend(module_name, params, storage, name, nofail=nofail)
        self.backend_instances[name] = backend
        return backend

    def unload_backends(self, names: str | List[str] | None = None) -> Dict[str, Module]:
        """
        Unload backends.

        :param names: if specified, only unload that backends
        :type names: :class:`list`
        """
        unloaded = {}
        if isinstance(names, str):
            names = [names]
        elif names is None:
            names = list(self.backend_instances.keys())

        for name in names:
            backend = self.backend_instances.pop(name)
            with backend:
                backend.deinit()
            unloaded[backend.name] = backend

        return unloaded

    def __getitem__(self, name: str) -> Module:
        """
        Alias for :func:`WoobBase.get_backend`.
        """
        return self.get_backend(name)

    def get_backend(self, name: str, **kwargs) -> Module:
        """
        Get a backend from its name.

        :param name: name of backend to get
        :type name: str
        :param default: if specified, get this value when the backend is not found
        :type default: whatever you want
        :raises: :class:`KeyError` if not found.
        """
        try:
            return self.backend_instances[name]
        except KeyError:
            if 'default' in kwargs:
                return kwargs['default']
            else:
                raise

    def count_backends(self) -> int:
        """
        Get number of loaded backends.
        """
        return len(self.backend_instances)

    def iter_backends(
        self,
        caps: List[Capability] | None = None,
        module: str | None = None
    ) -> Iterator[Module]:
        """
        Iter on each backends.

        Note: each backend is locked when it is returned.

        :param caps: optional list of capabilities to select backends
        :type caps: tuple[:class:`woob.capabilities.base.Capability`]
        :param module: optional name of module
        :type module: :class:`str`
        :rtype: iter[:class:`woob.tools.backend.Module`]
        """
        for _, backend in sorted(self.backend_instances.items()):
            if (caps is None or backend.has_caps(caps)) and \
               (module is None or backend.NAME == module):
                with backend:
                    yield backend

    def __getattr__(self, name: str) -> Callable[..., BackendsCall]:
        def caller(*args, **kwargs):
            return self.do(name, *args, **kwargs)
        return caller

    def do(self, function: Callable | str, *args, **kwargs) -> BackendsCall:
        r"""
        Do calls on loaded backends with specified arguments, in separated
        threads.

        This function has two modes:

        - If *function* is a string, it calls the method with this name on
          each backends with the specified arguments;
        - If *function* is a callable, it calls it in a separated thread with
          the locked backend instance at first arguments, and \*args and
          \*\*kwargs.

        :param function: backend's method name, or a callable object
        :type function: :class:`str`
        :param backends: list of backends to iterate on
        :type backends: list[:class:`str`]
        :param caps: iterate on backends which implement this caps
        :type caps: list[:class:`woob.capabilities.base.Capability`]
        :rtype: A :class:`woob.core.bcall.BackendsCall` object (iterable)
        """
        backends = list(self.backend_instances.values())
        _backends = kwargs.pop('backends', None)
        if _backends is not None:
            if isinstance(_backends, Module):
                backends = [_backends]
            elif isinstance(_backends, str):
                if len(_backends) > 0:
                    try:
                        backends = [self.backend_instances[_backends]]
                    except (ValueError, KeyError):
                        backends = []
            elif isinstance(_backends, (list, tuple, set)):
                backends = []
                for backend in _backends:
                    if isinstance(backend, str):
                        try:
                            backends.append(self.backend_instances[backend])
                        except (ValueError, KeyError):
                            pass
                    else:
                        backends.append(backend)
            else:
                self.logger.warning('The "backends" value isn\'t supported: %r', _backends)

        if 'caps' in kwargs:
            caps = kwargs.pop('caps')
            backends = [backend for backend in backends if backend.has_caps(caps)]

        # The return value MUST BE the BackendsCall instance. Please never iterate
        # here on this object, because caller might want to use other methods, like
        # wait() on callback_thread().
        # Thanks a lot.
        return BackendsCall(backends, function, *args, **kwargs)

    def schedule(self, interval: int, function: Callable, *args) -> int | None:
        """
        Schedule an event.

        :param interval: delay before calling the function
        :type interval: int
        :param function: function to call
        :type function: callabale
        :param args: arguments to give to function
        :returns: an event identificator
        """
        return self.scheduler.schedule(interval, function, *args)

    def repeat(self, interval: int, function: Callable, *args) -> int | None:
        """
        Repeat a call to a function

        :param interval: interval between two calls
        :type interval: int
        :param function: function to call
        :type function: callable
        :param args: arguments to give to function
        :returns: an event identificator
        """
        return self.scheduler.repeat(interval, function, *args)

    def cancel(self, ev: int) -> bool:
        """
        Cancel an event

        :param ev: the event identificator
        """
        return self.scheduler.cancel(ev)

    def want_stop(self) -> None:
        """
        Plan to stop the scheduler.
        """
        return self.scheduler.want_stop()

    def loop(self):
        """
        Run the scheduler loop
        """
        return self.scheduler.run()

    def load_or_install_module(self, module_name: str) -> Module:
        """ Load a backend, but can't install it """
        return self.modules_loader.get_or_load_module(module_name)


class Woob(WoobBase):
    """
    The main class of Woob, used to manage backends, modules repositories and
    call methods on all loaded backends.

    :param workdir: optional parameter to set path of the working directory
    :type workdir: str
    :param datadir: optional parameter to set path of the data directory
    :type datadir: str
    :param backends_filename: name of the *backends* file, where configuration of
                              backends is stored
    :type backends_filename: str
    :param storage: provide a storage where backends can save data
    :type storage: :class:`woob.tools.storage.IStorage`
    """
    BACKENDS_FILENAME = 'backends'

    def __init__(
        self,
        workdir: str | None = None,
        datadir: str | None = None,
        backends_filename: str | None = None,
        scheduler: IScheduler | None = None,
        storage: IStorage | None = None
    ):
        # Create WORKDIR
        xdg_config = Path(os.environ.get("XDG_CONFIG_HOME") or Path.home() / ".config")

        workdir_paths = [
            workdir,
            os.environ.get("WOOB_WORKDIR"),
            os.environ.get("WEBOOB_WORKDIR"),
        ]
        workdir = self._get_working_dir(workdir_paths, xdg_config / "woob")
        self.workdir = os.path.realpath(workdir)
        self._create_dir(self.workdir)

        # Create DATADIR
        xdg_data = Path(os.environ.get("XDG_DATA_HOME") or Path.home() / ".local" / "share")
        datadir_paths = [
            datadir,
            os.environ.get("WOOB_DATADIR"),
            os.environ.get("WOOB_WORKDIR"),
            os.environ.get("WEBOOB_DATADIR"),
            os.environ.get("WEBOOB_WORKDIR"),
        ]
        datadir = self._get_working_dir(datadir_paths, xdg_data / "woob")
        datadir = os.path.realpath(datadir)
        self._create_dir(datadir)

        # Modules management
        self.repositories = Repositories(workdir, datadir, __version__)

        # Backend instances config
        if not backends_filename:
            backends_filename = (
                os.environ.get('WOOB_BACKENDS')
                or os.environ.get('WEBOOB_BACKENDS')
                or os.path.join(self.workdir, self.BACKENDS_FILENAME)
            )
        elif not backends_filename.startswith('/'):
            backends_filename = os.path.join(self.workdir, backends_filename)
        self.backends_config: BackendsConfig = BackendsConfig(backends_filename)

        super().__init__(modules_path=None, scheduler=scheduler, storage=storage)

    def build_modules_loader(self) -> RepositoryModulesLoader:
        """
        Build the module loader for the current application.

        :rtype: ModulesLoader
        """
        return RepositoryModulesLoader(self.repositories)

    def _get_working_dir(self, priority_dirs: List[str | None], user_dir: str | Path) -> str:
        for directory in priority_dirs:
            if directory:
                return str(directory)

        return str(user_dir)

    def _create_dir(self, name: str):
        if not os.path.exists(name):
            os.makedirs(name)
        elif not os.path.isdir(name):
            self.logger.error('"%s" is not a directory', name)

    def update(self, progress: IProgress = PrintProgress()):
        """
        Update modules from repositories.

        :param progress: object notified when there is a progress
        :type progress: IProgress
        """
        self.repositories.update(progress)

        modules_to_check = set([module_name for _, module_name, _ in self.backends_config.iter_backends()])
        for module_name in modules_to_check:
            minfo = self.repositories.get_module_info(module_name)
            if minfo and not minfo.is_installed():
                self.repositories.install(minfo, progress)

    def build_backend(self,
                      module_name: str,
                      params: Dict[str, str] | None = None,
                      storage: IStorage | None = None,
                      name: str | None = None,
                      nofail: bool = False,
                      logger: logging.Logger | None = None) -> Module:
        """
        Create a single backend which is not listed in configuration.

        :param module_name: name of module
        :type module_name: str
        :param params: parameters to give to backend
        :type params: :class:`dict`
        :param storage: storage to use
        :type storage: :class:`woob.tools.storage.IStorage`
        :param name: name of backend
        :type name: :class:`str`
        :rtype: :class:`woob.tools.backend.Module`
        :param nofail: if true, this call can't fail
        :type nofail: :class:`bool`
        """
        minfo = self.repositories.get_module_info(module_name)
        if minfo is None:
            raise ModuleLoadError(module_name, 'Module does not exist.')

        if not minfo.is_installed():
            self.repositories.install(minfo)

        return super().build_backend(module_name, params, storage, name, nofail, logger)

    def load_backends(self,
                      caps: List[Capability | str] | None = None,
                      names: List[str] | None = None,
                      modules: List[str] | None = None,
                      exclude: List[str] | None= None,
                      storage: IStorage | None = None,
                      errors: List['Woob.LoadError'] | None = None) -> Dict[str, Module]:
        """
        Load backends listed in config file.

        :param caps: load backends which implement all of specified caps
        :type caps: tuple[:class:`woob.capabilities.base.Capability`]
        :param names: load backends in list
        :type names: tuple[:class:`str`]
        :param modules: load backends which module is in list
        :type modules: tuple[:class:`str`]
        :param exclude: do not load backends in list
        :type exclude: tuple[:class:`str`]
        :param storage: use this storage if specified
        :type storage: :class:`woob.tools.storage.IStorage`
        :param errors: if specified, store every errors in this list
        :type errors: list[:class:`LoadError`]
        :returns: loaded backends
        :rtype: dict[:class:`str`, :class:`woob.tools.backend.Module`]
        """
        loaded = {}
        if storage is None:
            storage = self.storage

        if not self.repositories.check_repositories():
            self.logger.error('Repositories are not consistent with the sources.list')
            raise VersionsMismatchError('Versions mismatch, please run "woob config update"')

        for backend_name, module_name, params in self.backends_config.iter_backends():
            if '_enabled' in params and not params['_enabled'].lower() in ('1', 'y', 'true', 'on', 'yes') or \
               names is not None and backend_name not in names or \
               modules is not None and module_name not in modules or \
               exclude is not None and backend_name in exclude:
                continue

            minfo = self.repositories.get_module_info(module_name)
            if minfo is None:
                self.logger.warning('Backend "%s" is referenced in %s but was not found. '
                                    'Perhaps a missing repository or a removed module?', module_name, self.backends_config.confpath)
                continue

            if caps is not None and not minfo.has_caps(caps):
                continue

            if not minfo.is_installed():
                self.repositories.install(minfo)

            module = None
            try:
                module = self.modules_loader.get_or_load_module(module_name)
            except ModuleLoadError as e:
                self.logger.error('Unable to load module "%s": %s', module_name, e)
                continue

            if backend_name in self.backend_instances:
                self.logger.warning('Oops, the backend "%s" is already loaded. Unload it before reloading...', backend_name)
                self.unload_backends(backend_name)

            try:
                backend_instance = module.create_instance(self, backend_name, params, storage)
            except Module.ConfigError as e:
                if errors is not None:
                    errors.append(self.LoadError(backend_name, str(e)))
            else:
                self.backend_instances[backend_name] = loaded[backend_name] = backend_instance
        return loaded

    def load_or_install_module(self, module_name: str) -> Module:
        """ Load a backend, and install it if not done before """
        try:
            return self.modules_loader.get_or_load_module(module_name)
        except ModuleLoadError:
            self.repositories.install(module_name)
            return self.modules_loader.get_or_load_module(module_name)



# Copyright(C) 2010-2011  Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import datetime
import hashlib
import shlex
import subprocess
from tempfile import NamedTemporaryFile

from lxml import etree

from woob.core import CallErrors
from woob.capabilities.base import empty
from woob.capabilities.messages import CapMessages, Message, Thread
from woob.capabilities.account import CapAccount
from woob.capabilities.contact import CapContact
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter
from woob.tools.html import html2text


__all__ = ['AppMsg']


class AtomFormatter(IFormatter):
    MANDATORY_FIELDS = ('title', 'date', 'sender', 'content')

    def _format_date(self, dt):
        return dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    def start_format(self, **kwargs):
        gen_time = datetime.datetime.utcnow()

        self.output('<?xml version="1.0" encoding="utf-8"?>')
        self.output('<feed xmlns="http://www.w3.org/2005/Atom" xmlns:dc="http://purl.org/dc/elements/1.1/">')
        self.output('  <title type="text">Atom feed by Woob</title>')  # TODO : get backend name
        self.output('  <updated>%s</updated>' % self._format_date(gen_time))

        m = hashlib.new("md5")  # nosec
        m.update(self._format_date(gen_time).encode("ascii"))
        self.output("  <id>urn:md5:%s</id>" % m.hexdigest())

    def format_obj(self, obj, alias):
        elem = etree.Element("entry")

        title = etree.SubElement(elem, "title")
        title.text = obj.title

        id = etree.SubElement(elem, "id")
        m = hashlib.new("md5")  # nosec
        m.update(obj.content.encode("utf8"))
        id.text = "urn:md5:%s" % m.hexdigest()

        link = etree.SubElement(elem, "link")
        link.attrib["href"] = obj.url
        link.attrib["title"] = obj.title
        link.attrib["type"] = "text/html"

        author = etree.SubElement(elem, "author")
        name = etree.SubElement(author, "name")
        if obj.sender:
            name.text = obj.sender
        else:
            name.text = obj.backend

        date = etree.SubElement(elem, "updated")
        date.text = self._format_date(obj.date)

        content = etree.SubElement(elem, 'content')
        content.text = obj.content
        content.attrib["type"] = "html"

        return etree.tostring(elem, pretty_print=True, encoding="unicode")

    def flush(self):
        self.output("</feed>")


class XHtmlFormatter(IFormatter):
    MANDATORY_FIELDS = ('title', 'date', 'sender', 'signature', 'content')

    def format_obj(self, obj, alias):
        result  = "<div>\n"
        result += "<h1>%s</h1>" % (obj.title)
        result += "<dl>"
        result += "<dt>Date</dt><dd>%s</dd>" % (obj.date.strftime('%Y-%m-%d %H:%M'))
        result += "<dt>Sender</dt><dd>%s</dd>" % (obj.sender)
        result += "<dt>Signature</dt><dd>%s</dd>" % (obj.signature)
        result += "</dl>"
        result += "<div>%s</div>" % (obj.content)
        result += "</div>\n"
        return result


class MessageFormatter(IFormatter):
    MANDATORY_FIELDS = ('title', 'date', 'sender', 'signature', 'content')

    def format_obj(self, obj, alias):
        result = '%sTitle:%s %s\n' % (self.BOLD,
                                       self.NC, obj.title)
        result += '%sDate:%s %s\n' % (self.BOLD,
                                       self.NC, obj.date.strftime('%Y-%m-%d %H:%M'))
        result += '%sFrom:%s %s\n' % (self.BOLD,
                                       self.NC, obj.sender)
        if hasattr(obj, 'receivers') and obj.receivers:
            result += '%sTo:%s %s\n' % (self.BOLD,
                                         self.NC,
                                         ', '.join(obj.receivers))

        if obj.flags & Message.IS_HTML:
            content = html2text(obj.content)
        else:
            content = obj.content

        result += '\n%s' % content

        if obj.signature:
            if obj.flags & Message.IS_HTML:
                signature = html2text(obj.signature)
            else:
                signature = obj.signature

            result += '\n-- \n%s' % signature
        return result


class MessagesListFormatter(IFormatter):
    MANDATORY_FIELDS = ()
    count = 0
    _list_messages = False

    def flush(self):
        self.count = 0

    def format_obj(self, obj, alias):
        if not self._list_messages:
            return self.format_dict_thread(obj, alias)
        else:
            return self.format_dict_messages(obj, alias)

    def format_dict_thread(self, obj, alias):
        self.count += 1
        if self.interactive:
            result = '%s* (%d) %s (%s)%s' % (self.BOLD,
                                              self.count,
                                              obj.title, obj.backend,
                                              self.NC)
        else:
            result = '%s* (%s) %s%s' % (self.BOLD, obj.id,
                                         obj.title,
                                         self.NC)
        if obj.date:
            result += '\n             %s' % obj.date
        return result

    def format_dict_messages(self, obj, alias):
        if obj.flags == Thread.IS_THREADS:
            depth = 0
        else:
            depth = -1

        result = self.format_message(obj.backend, obj.root, depth)
        return result

    def format_message(self, backend, message, depth=0):
        if not message:
            return ''
        self.count += 1

        flags = '['
        if message.flags & message.IS_UNREAD:
            flags += 'N'
        else:
            flags += '-'
        if message.flags & message.IS_NOT_RECEIVED:
            flags += 'U'
        elif message.flags & message.IS_RECEIVED:
            flags += 'R'
        else:
            flags += '-'
        flags += ']'

        if self.interactive:
            result = '%s%s* (%d)%s %s <%s> %s (%s)\n' % (depth * '  ',
                                                          self.BOLD,
                                                          self.count,
                                                          self.NC,
                                                          flags,
                                                          message.sender,
                                                          message.title,
                                                          backend)
        else:
            result = '%s%s* (%s.%s@%s)%s %s <%s> %s\n' % (depth * '  ',
                                                           self.BOLD,
                                                           message.thread.id,
                                                           message.id,
                                                           backend,
                                                           self.NC,
                                                           flags,
                                                           message.sender,
                                                           message.title)
        if message.children:
            if depth >= 0:
                depth += 1
            for m in message.children:
                result += self.format_message(backend, m, depth)
        return result


class ProfileFormatter(IFormatter):
    def flush(self):
        pass

    def format_obj(self, obj, alias=None):
        return obj.get_text()


class AppMsg(ReplApplication):
    APPNAME = 'msg'
    OLD_APPNAME = 'boobmsg'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Christophe Benz'
    DESCRIPTION = "Console application allowing to send messages on various websites and " \
                  "to display message threads and contents."
    SHORT_DESCRIPTION = "send and receive message threads"
    CAPS = CapMessages
    EXTRA_FORMATTERS = {'msglist':  MessagesListFormatter,
                        'msg':      MessageFormatter,
                        'xhtml':    XHtmlFormatter,
                        'atom':     AtomFormatter,
                        'profile' : ProfileFormatter,
                       }
    COMMANDS_FORMATTERS = {'list':          'msglist',
                           'show':          'msg',
                           'export_thread': 'msg',
                           'export_all':    'msg',
                           'ls':            'msglist',
                           'profile':       'profile',
                          }

    def add_application_options(self, group):
        group.add_option('-E', '--accept-empty',  action='store_true',
                         help='Send messages with an empty body.')
        group.add_option('-t', '--title', action='store',
                         help='For the "post" command, set a title to message',
                         type='string', dest='title')

    def load_default_backends(self):
        self.load_backends(CapMessages, storage=self.create_storage())

    def main(self, argv):
        self.load_config()
        return super().main(argv)

    def do_status(self, line):
        """
        status

        Display status information about a backend.
        """
        if len(line) > 0:
            backend_name = line
        else:
            backend_name = None

        results = {}
        for field in self.do('get_account_status',
                             backends=backend_name,
                             caps=CapAccount):
            if field.backend in results:
                results[field.backend].append(field)
            else:
                results[field.backend] = [field]

        for name, fields in results.items():
            print(':: %s ::' % name)
            for f in fields:
                if f.flags & f.FIELD_HTML:
                    value = html2text(f.value)
                else:
                    value = f.value
                print('%s: %s' % (f.label, value))
            print('')

    def do_post(self, line):
        """
        post RECEIVER@BACKEND[,RECEIVER@BACKEND[...]] [TEXT]

        Post a message to the specified receivers.
        Multiple receivers are separated by a comma.

        If no text is supplied on command line, the content of message is read on stdin.
        """
        receivers, text = self.parse_command_args(line, 2, 1)
        if text is None:
            text = self.acquire_input()

        if not self.options.accept_empty and not text.strip():
            self.logger.warning('The message body is empty, use option --accept_empty to send empty messages')
            return

        for receiver in receivers.strip().split(','):
            receiver, backend_name = self.parse_id(receiver.strip(),
                                                   unique_backend=True)
            if not backend_name and len(self.enabled_backends) > 1:
                self.logger.warning('No backend specified for receiver "%s": message will be sent with all the '
                    'enabled backends (%s)' % (receiver,
                    ','.join(backend.name for backend in self.enabled_backends)))

            if '.' in receiver:
                # It's a reply
                thread_id, parent_id = receiver.rsplit('.', 1)
            else:
                # It's an original message
                thread_id = receiver
                parent_id = None
                try:
                    thread_id = self.threads[int(thread_id) - 1].id
                except (IndexError,ValueError):
                    pass

            thread = Thread(thread_id)
            message = Message(thread,
                              0,
                              title=self.options.title,
                              parent=Message(thread, parent_id) if parent_id else None,
                              content=text)

            try:
                self.do('post_message', message, backends=backend_name).wait()
            except CallErrors as errors:
                self.bcall_errors_handler(errors)
            else:
                if self.interactive:
                    print('Message sent sucessfully to %s' % receiver)

    threads = []
    messages = []

    @defaultcount(10)
    def do_list(self, arg):
        """
        list

        Display all threads.
        """
        if len(arg) > 0:
            try:
                thread = self.threads[int(arg) - 1]
            except (IndexError, ValueError):
                id, backend_name = self.parse_id(arg)
            else:
                id = thread.id
                backend_name = thread.backend

            self.messages = []
            cmd = self.do('get_thread', id, backends=backend_name)
            self.formatter._list_messages = True
        else:
            self.threads = []
            cmd = self.do('iter_threads')
            self.formatter._list_messages = False

        self.start_format()
        for thread in cmd:
            if not thread:
                continue
            if len(arg) > 0:
                if not thread.root:
                    thread, = self.do("fillobj", thread, ("root",), backends=thread.backend)
                if thread.root:
                    thread.root, = self.do("fillobj", thread.root, ("children",), backends=thread.backend)

                for m in thread.iter_all_messages():
                    if not m.backend:
                        m.backend = thread.backend
                    self.messages.append(m)
            else:
                thread, = self.do("fillobj", thread, ("title", "date"), backends=thread.backend)
                self.threads.append(thread)
            self.format(thread)

    def do_export_all(self, arg):
        """
        export_all

        Export All threads
        """

        def func(backend):
            for thread in backend.iter_threads():
                if not thread:
                    continue
                t = backend.fillobj(thread, None)
                for msg in t.iter_all_messages():
                    yield msg

        self.start_format()
        for msg in self.do(func):
            self.format(msg)

    def do_export_thread(self, arg):
        """
        export_thread ID

        Export the thread identified by ID
        """
        try:
            thread = self.threads[int(arg) - 1]
        except (IndexError, ValueError):
            _id, backend_name = self.parse_id(arg)
        else:
            _id = thread.id
            backend_name = thread.backend

        cmd = self.do('get_thread', _id, backends=backend_name)
        self.start_format()
        for thread in cmd:
            if thread is not None:
                thread, = self.do('fillobj', thread, None, backends=thread.backend)
                for msg in thread.iter_all_messages():
                    msg, = self.do('fillobj', msg, None, backends=thread.backend)
                    self.format(msg)

    def do_show(self, arg):
        """
        show MESSAGE

        Read a message
        """
        message = None
        if len(arg) == 0:
            print('Please give a message ID.', file=self.stderr)
            return 2

        try:
            message = self.messages[int(arg) - 1]
        except (IndexError, ValueError):
            # The message is not is the cache, we have now two cases:
            # 1) the user uses a number to get a thread in the cache
            # 2) the user gives a thread id
            try:
                thread = self.threads[int(arg) - 1]
                if not empty(thread.root):
                    message = thread.root
                else:
                    for thread in self.do('get_thread', thread.id, backends=thread.backend):
                        if thread is not None:
                            if not thread.root:
                                thread, = self.do('fillobj', thread, ('root',), backends=thread.backend)
                            message = thread.root
            except (IndexError, ValueError):
                _id, backend_name = self.parse_id(arg)
                for thread in self.do('get_thread', _id, backends=backend_name):
                    if thread is not None:
                        if not thread.root:
                            thread, = self.do('fillobj', thread, ('root',), backends=thread.backend)
                        message = thread.root
        if not empty(message):
            message, = self.do("fillobj", message, backends=message.backend)

            self.start_format()
            self.format(message)
            self.woob.do('set_message_read', message, backends=message.backend)
            return
        else:
            print('Message not found', file=self.stderr)
            return 3

    def do_profile(self, id):
        """
        profile ID

        Display a profile
        """
        _id, backend_name = self.parse_id(id, unique_backend=True)

        found = 0
        for contact in self.do('get_contact', _id, backends=backend_name, caps=CapContact):
            if contact:
                self.format(contact)
                found = 1

        if not found:
            self.logger.error('Profile not found')

    def do_photos(self, id):
        """
        photos ID

        Display photos of a profile
        """
        photo_cmd = self.config.get('photo_viewer')
        if photo_cmd is None:
            print("Configuration error: photo_viewer is undefined", file=self.stderr)
            return

        photo_cmd = shlex.split(photo_cmd)
        if photo_cmd[-1] == '%s':
            del photo_cmd

        _id, backend_name = self.parse_id(id, unique_backend=True)

        found = 0
        for contact in self.do('get_contact', _id, backends=backend_name):
            if contact:
                # Write photo to temporary files
                tmp_files = []
                for photo in contact.photos.values():
                    suffix = '.jpg'
                    if '.' in photo.url.split('/')[-1]:
                        suffix = '.%s' % photo.url.split('/')[-1].split('.')[-1]
                    f = NamedTemporaryFile(suffix=suffix)

                    photo = self.woob[contact.backend].fillobj(photo, 'data')
                    f.write(photo.data)
                    tmp_files.append(f)

                subprocess.call(photo_cmd + [file.name for file in tmp_files])
                found = 1

        if not found:
            self.logger.error('Profile not found')



# Copyright(C) 2013 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, Field, StringField, UserError, Enum
from .date import DateField


class Event(BaseObject):
    date = DateField('Date')
    activity = StringField('Activity')
    location = StringField('Location')

    def __repr__(self):
        return '<Event date=%r activity=%r location=%r>' % (self.date, self.activity, self.location)


class ParcelState(Enum):
    UNKNOWN = 0
    PLANNED = 1
    IN_TRANSIT = 2
    ARRIVED = 3


class Parcel(BaseObject):
    STATUS_UNKNOWN = ParcelState.UNKNOWN
    STATUS_PLANNED = ParcelState.PLANNED
    STATUS_IN_TRANSIT = ParcelState.IN_TRANSIT
    STATUS_ARRIVED = ParcelState.ARRIVED

    arrival = DateField('Scheduled arrival date')
    status = Field('Status of parcel', int, default=STATUS_UNKNOWN)
    info = StringField('Information about parcel status')
    history = Field('History', list)


class CapParcel(Capability):
    def get_parcel_tracking(self, id):
        """
        Get information abouut a parcel.

        :param id: ID of the parcel
        :type id: :class:`str`
        :rtype: :class:`Parcel`
        :raises: :class:`ParcelNotFound`
        """

        raise NotImplementedError()


class ParcelNotFound(UserError):
    """
    Raised when a parcell is not found.
    It can be an user error, or an expired parcel
    """

    def __init__(self, msg='Parcel not found'):
        super(ParcelNotFound, self).__init__(msg)



# Copyright(C) 2010-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
from unittest import TestCase
from unittest.case import SkipTest

from woob.capabilities.base import empty
from woob.core import Woob


__all__ = ['BackendTest', 'SkipTest', 'skip_without_config']


class BackendTest(TestCase):
    MODULE = None

    def __init__(self, *args, **kwargs):
        super(BackendTest, self).__init__(*args, **kwargs)

        self.backends = {}
        self.backend_instance = None
        self.backend = None
        self.woob = Woob()

        # Skip tests when passwords are missing
        self.woob.requests.register('login', self.login_cb)

        if self.woob.load_backends(modules=[self.MODULE]):
            # provide the tests with all available backends
            self.backends = self.woob.backend_instances

    def login_cb(self, backend_name, value):
        raise SkipTest('missing config \'%s\' is required for this test' % value.label)

    def run(self, result):
        """
        Call the parent run() for each backend instance.
        Skip the test if we have no backends.
        """
        try:
            if not len(self.backends):
                self.backend = self.woob.build_backend(self.MODULE, nofail=True)
                super().run(result)
            else:
                # Run for all backend
                for backend_instance in self.backends.keys():
                    print(backend_instance)
                    self.backend = self.backends[backend_instance]
                    super().run(result)
        finally:
            self.woob.deinit()

    def shortDescription(self):
        """
        Generate a description with the backend instance name.
        """
        # do not use TestCase.shortDescription as it returns None
        return '%s [%s]' % (str(self), self.backend_instance)

    def is_backend_configured(self):
        """
        Check if the backend is in the user configuration file
        """
        return self.woob.backends_config.backend_exists(self.backend.config.instname)

    def assertNotEmpty(self, obj, *args):
        """
        Assert an object is neither `empty` in the BaseObject parlance.

        `obj` should not be `None`, `NotLoaded`, or `NotAvailable`.
        """
        self.assertFalse(empty(obj), *args)


def skip_without_config(*keys):
    """Decorator to skip a test if backend config is missing

    :param keys: if any of these keys is missing in backend config, skip test. Can be empty.
    """

    for key in keys:
        if callable(key):
            raise TypeError('skip_without_config() must be called with arguments')

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            config = self.backend.config
            if not self.is_backend_configured():
                raise SkipTest('a backend must be declared in configuration for this test')
            for key in keys:
                if not config[key].get():
                    raise SkipTest('config key %r is required for this test' %
                                   key)

            return func(self, *args, **kwargs)
        return wrapper
    return decorator



# Copyright(C) 2012-2016 woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['CacheMixin']


class CacheEntry:
    def __init__(self, response):
        self.response = response
        self.etag = response.headers.get('ETag')
        self.last_modified = response.headers.get('Last-Modified')

    def has_cache_key(self):
        return (self.etag or self.last_modified)

    def update_request(self, request):
        if self.last_modified:
            request.headers['If-Modified-Since'] = self.last_modified
        if self.etag:
            request.headers['If-None-Match'] = self.etag


class CacheMixin:
    """Mixin to inherit in a Browser"""

    cache_is_updatable = True

    """Whether the cache is updatable

    If `False`, once a request has been successfully executed, its response
    will always be returned.

    If `True`, the `ETag` and `Last-Modified` of the response will be
    stored along with the cache. When the request is re-executed, instead
    of simply returning the previous response, the server is queried to
    check if a newer version of the page exists.
    If a newer page exists, it is returned instead and overwrites the
    obsolete page in the cache.
    """

    def __init__(self, *args, **kwargs):
        super(CacheMixin, self).__init__(*args, **kwargs)

        self.cache = {}

        """Cache store object

        To limit the size of the cache, a :class:`woob.tools.lrudict.LimitedLRUDict`
        instance can be used.
        """

    def make_cache_key(self, request):
        """Make a key for the cache corresponding to the request."""

        body = getattr(request, 'body', None)
        headers = tuple(request.headers.values())
        return (request.method, request.url, body, headers)

    def open_with_cache(self, url, **kwargs):
        """Perform a request using the cache if possible."""
        request = self.build_request(url, **kwargs)

        key = self.make_cache_key(request)
        if key in self.cache:
            if not self.cache_is_updatable:
                self.logger.debug('cache HIT for %r', request.url)
                return self.cache[key].response
            else:
                self.cache[key].update_request(request)

        response = super(CacheMixin, self).open(request, **kwargs)
        if response.status_code == 304:
            self.logger.debug('cache HIT for %r', request.url)
            return self.cache[key].response
        elif response.status_code == 200:
            entry = CacheEntry(response)
            if entry.has_cache_key():
                self.logger.debug('storing %r response in cache', request.url)
                self.cache[key] = entry

        self.logger.debug('cache MISS for %r', request.url)
        return response



from threading import Lock, Event

from woob.capabilities.captcha import CapCaptchaSolver
from woob.core import Woob

__all__ = ['CaptchaMixin']


class CaptchaMixin:
    def __init__(self, *args, **kwargs):
        super(CaptchaMixin, self).__init__(*args, **kwargs)
        self.captcha_woob = Woob()
        self.captcha_woob.load_backends(caps=[CapCaptchaSolver])

    def solve_captcha(self, job, backend):
        def call_solver(solver_backend, job):
            with lock:
                if solved.is_set():
                    solver_backend.logger.info('already solved, ignoring')
                    return

                ret = solver_backend.solve_captcha_blocking(job)
                if ret:
                    solver_backend.logger.info('backend solved job')
                    backend.config['captcha_response'].set(ret.solution)
                    solved.set()

        def all_solvers_finished():
            if not solved.is_set():
                print('Error(%s): CAPTCHA could not be solved.' % backend.name, file=self.stderr)
            else:
                print('Info(%s): CAPTCHA was successfully solved. Please retry operation.' % backend.name, file=self.stderr)

        lock = Lock()
        solved = Event()

        bres = self.captcha_woob.do(call_solver, job)
        bres.callback_thread(None, None, all_solvers_finished)



# Copyright(C) 2013 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from woob.applications.torrent.torrent import TorrentInfoFormatter, TorrentListFormatter
from woob.applications.subtitles.subtitles import SubtitleInfoFormatter, SubtitleListFormatter
from woob.capabilities.torrent import CapTorrent, MagnetOnly
from woob.capabilities.cinema import CapCinema
from woob.capabilities.subtitle import CapSubtitle
from woob.capabilities.base import empty, NotAvailable
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.core import CallErrors


__all__ = ['AppCinema']

ROLE_LIST = ['actor', 'director', 'writer', 'composer', 'producer']
COUNTRY_LIST = ['us', 'fr', 'de', 'jp']


class MovieInfoFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'original_title', 'release_date',
                        'other_titles', 'duration', 'pitch', 'note', 'roles', 'country')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.original_title, self.NC)
        result += 'ID: %s\n' % obj.fullid
        if not empty(obj.release_date):
            result += 'Released: %s\n' % obj.release_date.strftime('%Y-%m-%d')
        result += 'Country: %s\n' % obj.country
        if not empty(obj.duration):
            result += 'Duration: %smin\n' % obj.duration
        result += 'Note: %s\n' % obj.note
        if not empty(obj.genres):
            result += '\n%sGenres%s\n' % (self.BOLD, self.NC)
            for g in obj.genres:
                result += ' * %s\n' % g
        if not empty(obj.roles):
            result += '\n%sRelated persons%s\n' % (self.BOLD, self.NC)
            for role, lpersons in obj.roles.items():
                result += ' -- %s\n' % role
                for person in lpersons:
                    result += '   * %s\n' % person[1]
        if not empty(obj.other_titles):
            result += '\n%sOther titles%s\n' % (self.BOLD, self.NC)
            for t in obj.other_titles:
                result += ' * %s\n' % t
        if not empty(obj.pitch):
            result += '\n%sStory%s\n' % (self.BOLD, self.NC)
            result += '%s' % obj.pitch
        return result


class MovieListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'original_title', 'short_description')

    def get_title(self, obj):
        return obj.original_title

    def get_description(self, obj):
        result = ''
        if not empty(obj.short_description):
            result = obj.short_description
        return result


class MovieReleasesFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'original_title', 'all_release_dates')

    def get_title(self, obj):
        return 'Releases of %s' % obj.original_title

    def get_description(self, obj):
        return '\n%s' % obj.all_release_dates


def yearsago(years, from_date=None):
    if from_date is None:
        from_date = datetime.now()
    try:
        return from_date.replace(year=from_date.year - years)
    except:
        # Must be 2/29
        assert from_date.month == 2 and from_date.day == 29
        return from_date.replace(month=2, day=28,
                                 year=from_date.year-years)


def num_years(begin, end=None):
    if end is None:
        end = datetime.now()
    num_years = int((end - begin).days / 365.25)
    if begin > yearsago(num_years, end):
        return num_years - 1
    else:
        return num_years


class PersonInfoFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'birth_date', 'birth_place', 'short_biography')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.name, self.NC)
        result += 'ID: %s\n' % obj.fullid
        if not empty(obj.real_name):
            result += 'Real name: %s\n' % obj.real_name
        if not empty(obj.birth_place):
            result += 'Birth place: %s\n' % obj.birth_place
        if not empty(obj.birth_date):
            result += 'Birth date: %s\n' % obj.birth_date.strftime('%Y-%m-%d')
            if not empty(obj.death_date):
                age = num_years(obj.birth_date, obj.death_date)
                result += 'Death date: %s at %s years old\n' % (obj.death_date.strftime('%Y-%m-%d'), age)
            else:
                age = num_years(obj.birth_date)
                result += 'Age: %s\n' % age
        if not empty(obj.gender):
            result += 'Gender: %s\n' % obj.gender
        if not empty(obj.nationality):
            result += 'Nationality: %s\n' % obj.nationality
        if not empty(obj.roles):
            result += '\n%sRelated movies%s\n' % (self.BOLD, self.NC)
            for role, lmovies in obj.roles.items():
                result += ' -- %s\n' % role
                for movie in lmovies:
                    result += '   * %s\n' % movie[1]
        if not empty(obj.short_biography):
            result += '\n%sShort biography%s\n' % (self.BOLD, self.NC)
            result += '%s' % obj.short_biography
        return result


class PersonListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'short_description')

    def get_title(self, obj):
        return obj.name

    def get_description(self, obj):
        result = ''
        if not empty(obj.short_description):
            result = obj.short_description
        return result


class PersonBiographyFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'biography')

    def get_title(self, obj):
        return 'Biography of %s' % obj.name

    def get_description(self, obj):
        result = '\n%s' % obj.biography
        return result


class AppCinema(ReplApplication):
    APPNAME = 'cinema'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2013-YEAR Julien Veyssier'
    DESCRIPTION = "Console application allowing to search for movies and persons on various cinema databases " \
                  ", list persons related to a movie, list movies related to a person and list common movies " \
                  "of two persons."
    SHORT_DESCRIPTION = "search movies and persons around cinema"
    CAPS = (CapCinema, CapTorrent, CapSubtitle)
    EXTRA_FORMATTERS = {'movie_list': MovieListFormatter,
                        'movie_info': MovieInfoFormatter,
                        'movie_releases': MovieReleasesFormatter,
                        'person_list': PersonListFormatter,
                        'person_info': PersonInfoFormatter,
                        'person_bio': PersonBiographyFormatter,
                        'torrent_list': TorrentListFormatter,
                        'torrent_info': TorrentInfoFormatter,
                        'subtitle_list': SubtitleListFormatter,
                        'subtitle_info': SubtitleInfoFormatter
                        }
    COMMANDS_FORMATTERS = {'search_movie':    'movie_list',
                           'info_movie':      'movie_info',
                           'search_person':   'person_list',
                           'info_person':     'person_info',
                           'casting':         'person_list',
                           'filmography':     'movie_list',
                           'biography':     'person_bio',
                           'releases':     'movie_releases',
                           'movies_in_common': 'movie_list',
                           'persons_in_common': 'person_list',
                           'search_torrent':    'torrent_list',
                           'search_movie_torrent':    'torrent_list',
                           'info_torrent':      'torrent_info',
                           'search_subtitle':    'subtitle_list',
                           'search_movie_subtitle':    'subtitle_list',
                           'info_subtitle':      'subtitle_info'
                           }

    def complete_filmography(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 3:
            return ROLE_LIST

    def complete_casting(self, text, line, *ignored):
        return self.complete_filmography(text, line, ignored)

    def do_movies_in_common(self, line):
        """
        movies_in_common  person_ID  person_ID

        Get the list of common movies between two persons.
        """
        id1, id2 = self.parse_command_args(line, 2, 1)

        person1 = self.get_object(id1, 'get_person', caps=CapCinema)
        if not person1:
            print('Person not found: %s' % id1, file=self.stderr)
            return 3
        person2 = self.get_object(id2, 'get_person', caps=CapCinema)
        if not person2:
            print('Person not found: %s' % id2, file=self.stderr)
            return 3

        initial_count = self.options.count
        self.options.count = None

        lid1 = []
        for id in self.do('iter_person_movies_ids', person1.id, caps=CapCinema):
            lid1.append(id)
        lid2 = []
        for id in self.do('iter_person_movies_ids', person2.id, caps=CapCinema):
            lid2.append(id)
        self.options.count = initial_count
        inter = list(set(lid1) & set(lid2))

        chrono_list = []
        for common in inter:
            movie = self.get_object(common, 'get_movie', caps=CapCinema)
            role1 = movie.get_roles_by_person_id(person1.id)
            if not role1:
                role1 = movie.get_roles_by_person_name(person1.name)
            role2 = movie.get_roles_by_person_id(person2.id)
            if not role2:
                role2 = movie.get_roles_by_person_name(person2.name)

            if (movie.release_date != NotAvailable):
                year = movie.release_date.year
            else:
                year = '????'
            movie.short_description = '(%s) %s as %s ; %s as %s'%(year, person1.name, ', '.join(role1), person2.name, ', '.join(role2))
            if movie:
                i = 0
                while (i<len(chrono_list) and movie.release_date != NotAvailable and
                      (chrono_list[i].release_date == NotAvailable or year > chrono_list[i].release_date.year)):
                    i += 1
                chrono_list.insert(i, movie)

        for movie in chrono_list:
            self.cached_format(movie)

    def do_persons_in_common(self, line):
        """
        persons_in_common  movie_ID  movie_ID

        Get the list of common persons between two movies.
        """
        id1, id2 = self.parse_command_args(line, 2, 1)

        movie1 = self.get_object(id1, 'get_movie', caps=CapCinema)
        if not movie1:
            print('Movie not found: %s' % id1, file=self.stderr)
            return 3
        movie2 = self.get_object(id2, 'get_movie', caps=CapCinema)
        if not movie2:
            print('Movie not found: %s' % id2, file=self.stderr)
            return 3

        initial_count = self.options.count
        self.options.count = None

        lid1 = []
        for id in self.do('iter_movie_persons_ids', movie1.id, caps=CapCinema):
            lid1.append(id)
        lid2 = []
        for id in self.do('iter_movie_persons_ids', movie2.id, caps=CapCinema):
            lid2.append(id)
        self.options.count = initial_count
        inter = list(set(lid1) & set(lid2))
        for common in inter:
            person = self.get_object(common, 'get_person', caps=CapCinema)
            role1 = person.get_roles_by_movie_id(movie1.id)
            if not role1:
                role1 = person.get_roles_by_movie_title(movie1.original_title)
            role2 = person.get_roles_by_movie_id(movie2.id)
            if not role2:
                role2 = person.get_roles_by_movie_title(movie2.original_title)
            person.short_description = '%s in %s ; %s in %s'%(', '.join(role1), movie1.original_title, ', '.join(role2), movie2.original_title)
            self.cached_format(person)

    def do_info_movie(self, id):
        """
        info_movie  movie_ID

        Get information about a movie.
        """
        movie = self.get_object(id, 'get_movie', caps=CapCinema)

        if not movie:
            print('Movie not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(movie)

    def do_info_person(self, id):
        """
        info_person  person_ID

        Get information about a person.
        """
        person = self.get_object(id, 'get_person', caps=CapCinema)

        if not person:
            print('Person not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(person)

    @defaultcount(10)
    def do_search_movie(self, pattern):
        """
        search_movie  [PATTERN]

        Search movies.
        """
        self.change_path(['search movies'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for movie in self.do('iter_movies', pattern=pattern, caps=CapCinema):
            self.cached_format(movie)

    @defaultcount(10)
    def do_search_person(self, pattern):
        """
        search_person  [PATTERN]

        Search persons.
        """
        self.change_path(['search persons'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for person in self.do('iter_persons', pattern=pattern, caps=CapCinema):
            self.cached_format(person)

    def do_casting(self, line):
        """
        casting  movie_ID  [ROLE]

        List persons related to a movie.
        If ROLE is given, filter by ROLE
        """
        movie_id, role = self.parse_command_args(line, 2, 1)

        movie = self.get_object(movie_id, 'get_movie', caps=CapCinema)
        if not movie:
            print('Movie not found: %s' % id, file=self.stderr)
            return 3

        for person in self.do('iter_movie_persons', movie.id, role, backends=movie.backend, caps=CapCinema):
            self.cached_format(person)

    def do_filmography(self, line):
        """
        filmography  person_ID  [ROLE]

        List movies of a person.
        If ROLE is given, filter by ROLE
        """
        person_id, role = self.parse_command_args(line, 2, 1)

        person = self.get_object(person_id, 'get_person', caps=CapCinema)
        if not person:
            print('Person not found: %s' % id, file=self.stderr)
            return 3

        for movie in self.do('iter_person_movies', person.id, role, backends=person.backend, caps=CapCinema):
            self.cached_format(movie)

    def do_biography(self, person_id):
        """
        biography  person_ID

        Show the complete biography of a person.
        """
        person = self.get_object(person_id, 'get_person', ('name', 'biography'), caps=CapCinema)
        if not person:
            print('Person not found: %s' % person_id, file=self.stderr)
            return 3

        self.start_format()
        self.format(person)

    def complete_releases(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()
        if len(args) == 3:
            return COUNTRY_LIST

    def do_releases(self, line):
        """
        releases  movie_ID [COUNTRY]

        Get releases dates of a movie.
        If COUNTRY is given, show release in this country.
        """
        id, country = self.parse_command_args(line, 2, 1)

        movie = self.get_object(id, 'get_movie', ('original_title'), caps=CapCinema)
        if not movie:
            print('Movie not found: %s' % id, file=self.stderr)
            return 3

        # i would like to clarify with fillobj but how could i fill the movie AND choose the country ?
        for release in self.do('get_movie_releases', movie.id, country, caps=CapCinema, backends=movie.backend):
            if not empty(release):
                movie.all_release_dates = '%s' % (release)
            else:
                print('Movie releases not found for %s' % movie.original_title, file=self.stderr)
                return 3
        self.start_format()
        self.format(movie)

    # ================== TORRENT ==================

    def complete_info_torrent(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info_torrent(self, id):
        """
        info_torrent ID

        Get information about a torrent.
        """

        torrent = self.get_object(id, 'get_torrent', caps=CapTorrent)
        if not torrent:
            print('Torrent not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(torrent)

    def complete_getfile_torrent(self, text, line, *ignored):
        args = line.split(' ', 2)
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_getfile_torrent(self, line):
        """
        getfile_torrent ID [FILENAME]

        Get the .torrent file.
        FILENAME is where to write the file. If FILENAME is '-',
        the file is written to stdout.
        """
        id, dest = self.parse_command_args(line, 2, 1)

        _id, backend_name = self.parse_id(id)

        if dest is None:
            dest = '%s.torrent' % _id

        try:
            for buf in self.do('get_torrent_file', _id, backends=backend_name, caps=CapTorrent):
                if buf:
                    if dest == '-':
                        print(buf)
                    else:
                        try:
                            with open(dest, 'wb') as f:
                                f.write(buf)
                        except IOError as e:
                            print('Unable to write .torrent in "%s": %s' % (dest, e), file=self.stderr)
                            return 1
                    return
        except CallErrors as errors:
            for backend, error, backtrace in errors:
                if isinstance(error, MagnetOnly):
                    print('Error(%s): No direct URL available, '
                          'please provide this magnet URL '
                          'to your client:\n%s' % (backend, error.magnet), file=self.stderr)
                    return 4
                else:
                    self.bcall_error_handler(backend, error, backtrace)

        print('Torrent "%s" not found' % id, file=self.stderr)
        return 3

    @defaultcount(10)
    def do_search_torrent(self, pattern):
        """
        search_torrent [PATTERN]

        Search torrents.
        """
        self.change_path(['search torrent'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for torrent in self.do('iter_torrents', pattern=pattern, caps=CapTorrent):
            self.cached_format(torrent)

    @defaultcount(10)
    def do_search_movie_torrent(self, id):
        """
        search_movie_torrent movie_ID

        Search torrents of movie_ID.
        """
        movie = self.get_object(id, 'get_movie', ('original_title'), caps=CapCinema)
        if not movie:
            print('Movie not found: %s' % id, file=self.stderr)
            return 3

        pattern = movie.original_title

        self.change_path(['search torrent'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for torrent in self.do('iter_torrents', pattern=pattern, caps=CapTorrent):
            self.cached_format(torrent)

    # ================== SUBTITLE ==================

    def complete_info_subtitle(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info_subtitle(self, id):
        """
        info_subtitle subtitle_ID

        Get information about a subtitle.
        """

        subtitle = self.get_object(id, 'get_subtitle', caps=CapCinema)
        if not subtitle:
            print('Subtitle not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(subtitle)

    def complete_getfile_subtitle(self, text, line, *ignored):
        args = line.split(' ', 2)
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_getfile_subtitle(self, line):
        """
        getfile_subtitle subtitle_ID [FILENAME]

        Get the subtitle or archive file.
        FILENAME is where to write the file. If FILENAME is '-',
        the file is written to stdout.
        """
        id, dest = self.parse_command_args(line, 2, 1)

        _id, backend_name = self.parse_id(id)

        if dest is None:
            dest = '%s' % _id

        for buf in self.do('get_subtitle_file', _id, backends=backend_name, caps=CapSubtitle):
            if buf:
                if dest == '-':
                    print(buf)
                else:
                    try:
                        with open(dest, 'w') as f:
                            f.write(buf)
                    except IOError as e:
                        print('Unable to write file in "%s": %s' % (dest, e), file=self.stderr)
                        return 1
                return

        print('Subtitle "%s" not found' % id, file=self.stderr)
        return 3

    @defaultcount(10)
    def do_search_subtitle(self, line):
        """
        search_subtitle language [PATTERN]

        Search subtitles.

        Language  Abbreviation
        ----------------------
        Arabic      ar          Esperanto   eo          Irish       ga          Russian     ru
        Afrikaans   af          Estonian    et          Italian     it          Serbian     sr
        Albanian    sq          Filipino    tl          Japanese    ja          Slovak      sk
        Armenian    hy          Finnish     fi          Kannada     kn          Slovenian   sl
        Azerbaijani az          French      fr          Korean      ko          Spanish     es
        Basque      eu          Galician    gl          Latin       la          Swahili     sw
        Belarusian  be          Georgian    ka          Latvian     lv          Swedish     sv
        Bengali     bn          German      de          Lithuanian  lt          Tamil       ta
        Bulgarian   bg          Greek       gr          Macedonian  mk          Telugu      te
        Catalan     ca          Gujarati    gu          Malay       ms          Thai        th
        Chinese     zh          Haitian     ht          Maltese     mt          Turkish     tr
        Croatian    hr          Hebrew      iw          Norwegian   no          Ukrainian   uk
        Czech       cz          Hindi       hi          Persian     fa          Urdu        ur
        Danish      da          Hungaric    hu          Polish      pl          Vietnamese  vi
        Dutch       nl          Icelandic   is          Portuguese  pt          Welsh       cy
        English     en          Indonesian  id          Romanian    ro          Yiddish     yi
        ----------------------
        """
        language, pattern = self.parse_command_args(line, 2, 1)
        self.change_path(['search subtitle'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for subtitle in self.do('iter_subtitles', language=language, pattern=pattern, caps=CapSubtitle):
            self.cached_format(subtitle)

    @defaultcount(10)
    def do_search_movie_subtitle(self, line):
        """
        search_movie_subtitle language movie_ID

        Search subtitles of movie_ID.

        Language  Abbreviation
        ----------------------
        Arabic      ar          Esperanto   eo          Irish       ga          Russian     ru
        Afrikaans   af          Estonian    et          Italian     it          Serbian     sr
        Albanian    sq          Filipino    tl          Japanese    ja          Slovak      sk
        Armenian    hy          Finnish     fi          Kannada     kn          Slovenian   sl
        Azerbaijani az          French      fr          Korean      ko          Spanish     es
        Basque      eu          Galician    gl          Latin       la          Swahili     sw
        Belarusian  be          Georgian    ka          Latvian     lv          Swedish     sv
        Bengali     bn          German      de          Lithuanian  lt          Tamil       ta
        Bulgarian   bg          Greek       gr          Macedonian  mk          Telugu      te
        Catalan     ca          Gujarati    gu          Malay       ms          Thai        th
        Chinese     zh          Haitian     ht          Maltese     mt          Turkish     tr
        Croatian    hr          Hebrew      iw          Norwegian   no          Ukrainian   uk
        Czech       cz          Hindi       hi          Persian     fa          Urdu        ur
        Danish      da          Hungaric    hu          Polish      pl          Vietnamese  vi
        Dutch       nl          Icelandic   is          Portuguese  pt          Welsh       cy
        English     en          Indonesian  id          Romanian    ro          Yiddish     yi
        ----------------------
        """
        language, id = self.parse_command_args(line, 2, 2)
        movie = self.get_object(id, 'get_movie', ('original_title'), caps=CapCinema)
        if not movie:
            print('Movie not found: %s' % id, file=self.stderr)
            return 3

        pattern = movie.original_title
        self.change_path(['search subtitle'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for subtitle in self.do('iter_subtitles', language=language, pattern=pattern, caps=CapSubtitle):
            self.cached_format(subtitle)



# Copyright(C) 2017  Vincent Ardisson
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from datetime import date

from woob.capabilities.base import empty, NotLoaded
from woob.capabilities.bank import CapTransfer
from woob.capabilities.bank.wealth import CapBankWealth
from woob.exceptions import NoAccountsException
from woob.tools.capabilities.bank.iban import is_iban_valid
from woob.tools.capabilities.bank.investments import is_isin_valid
from woob.tools.date import new_date


__all__ = ('BankStandardTest',)


def sign(n):
    if n < 0:
        return -1
    elif n > 0:
        return 1
    else:
        return 0


class BankStandardTest:
    """Mixin for simple tests on CapBank backends.

    This checks:
    * there are accounts
    * accounts have an id, a label and a balance
    * history is implemented (optional)
    * transactions have a date, a label and an amount
    * investments are implemented (optional)
    * investments have a label and a valuation
    * recipients are implemented (optional)
    * recipients have an id and a label
    """

    allow_notimplemented_history = False
    allow_notimplemented_coming = False
    allow_notimplemented_investments = False
    allow_notimplemented_pockets = False
    allow_notimplemented_market_orders = False
    allow_notimplemented_emitters = False
    allow_notimplemented_recipients = False

    def test_basic(self):
        try:
            accounts = list(self.backend.iter_accounts())
        except NoAccountsException:
            return

        self.assertTrue(accounts)

        for account in accounts:
            self.check_account(account)

            try:
                self.check_history(account)
            except NotImplementedError:
                self.assertTrue(self.allow_notimplemented_history, 'iter_history should not raise NotImplementedError')

            try:
                self.check_coming(account)
            except NotImplementedError:
                self.assertTrue(self.allow_notimplemented_coming, 'iter_coming should not raise NotImplementedError')

            try:
                self.check_investments(account)
            except NotImplementedError:
                self.assertTrue(self.allow_notimplemented_investments, 'iter_investment should not raise NotImplementedError')

            try:
                self.check_pockets(account)
            except NotImplementedError:
                self.assertTrue(self.allow_notimplemented_pockets, 'iter_pocket should not raise NotImplementedError')

            try:
                self.check_market_orders(account)
            except NotImplementedError:
                self.assertTrue(self.allow_notimplemented_market_orders, 'iter_market_orders should not raise NotImplementedError')

            try:
                self.check_recipients(account)
            except NotImplementedError:
                self.assertTrue(self.allow_notimplemented_recipients, 'iter_transfer_recipients should not raise NotImplementedError')

        try:
            self.check_emitters()
        except NotImplementedError:
            self.assertTrue(self.allow_notimplemented_emitters, 'iter_emitters should not raise NotImplementedError')

    def check_account(self, account):
        self.assertTrue(account.id, 'account %r has no id' % account)
        self.assertTrue(account.label, 'account %r has no label' % account)
        self.assertFalse(empty(account.balance) and empty(account.coming), 'account %r should have balance or coming' % account)
        self.assertTrue(account.type, 'account %r is untyped' % account)
        self.assertTrue(account.currency, 'account %r has no currency' % account)
        self.assertIsNot(account.number, NotLoaded, 'account %r number is not loaded' % account)
        if account.iban:
            self.assertTrue(is_iban_valid(account.iban), 'account %r IBAN is invalid: %r' % (account, account.iban))

        if account.type in (account.TYPE_LOAN,):
            self.assertLessEqual(account.balance, 0, 'loan %r should not have positive balance' % account)
        elif account.type == account.TYPE_CHECKING:
            self.assertTrue(account.iban, 'account %r has no IBAN' % account)
        elif account.type == account.TYPE_CARD:
            if not account.parent:
                self.backend.logger.warning('card account %r has no parent account', account)
            else:
                self.assertEqual(account.parent.type, account.TYPE_CHECKING, 'parent account of %r should have checking type' % account)

    def check_history(self, account):
        for tr in self.backend.iter_history(account):
            self.check_transaction(account, tr, False)

    def check_coming(self, account):
        for tr in self.backend.iter_coming(account):
            self.check_transaction(account, tr, True)

    def check_transaction(self, account, tr, coming):
        today = date.today()

        self.assertFalse(empty(tr.date), 'transaction %r has no debit date' % tr)
        if tr.amount != 0:
            self.assertTrue(tr.amount, 'transaction %r has no amount' % tr)
        self.assertFalse(empty(tr.raw) and empty(tr.label), 'transaction %r has no raw or label' % tr)

        if coming:
            self.assertGreaterEqual(new_date(tr.date), today, 'coming transaction %r should be in the future' % tr)
        else:
            self.assertLessEqual(new_date(tr.date), today, 'history transaction %r should be in the past' % tr)

        if tr.rdate:
            self.assertGreaterEqual(new_date(tr.date), new_date(tr.rdate), 'transaction %r rdate should be before date' % tr)
            self.assertLess(abs(tr.date.year - tr.rdate.year), 2, 'transaction %r date (%r) and rdate (%r) are too far away' % (tr, tr.date, tr.rdate))

        if tr.original_amount or tr.original_currency:
            self.assertTrue(tr.original_amount and tr.original_currency, 'transaction %r has missing foreign info' % tr)

        for inv in (tr.investments or []):
            self.assertTrue(inv.label, 'transaction %r investment %r has no label' % (tr, inv))
            self.assertTrue(inv.valuation, 'transaction %r investment %r has no valuation' % (tr, inv))

    def check_investments(self, account):
        if not isinstance(self.backend, CapBankWealth):
            return

        total = 0
        for inv in self.backend.iter_investment(account):
            self.check_investment(account, inv)
            if not empty(inv.valuation):
                total += inv.valuation

        if total:
            self.assertEqual(total, account.balance, 'investments total (%s) is different than account balance (%s)' % (total, account.balance))

    def check_investment(self, account, inv):
        self.assertTrue(inv.label, 'investment %r has no label' % inv)
        self.assertFalse(empty(inv.valuation), 'investment %r has no valuation' % inv)
        if inv.code and not inv.code.startswith('XX-'):
            self.assertTrue(inv.code_type, 'investment %r has code but no code type' % inv)
        if inv.code_type == inv.CODE_TYPE_ISIN:
            self.assertTrue(is_isin_valid(inv.code), 'investment %r has invalid ISIN: %r' % (inv, inv.code))
        if not empty(inv.portfolio_share):
            self.assertTrue(0 < inv.portfolio_share <= 1, 'investment %r has invalid portfolio_share' % inv)

    def check_pockets(self, account):
        if not isinstance(self.backend, CapBankWealth):
            return
        for pocket in self.backend.iter_pocket(account):
            self.check_pocket(account, pocket)

    def check_pocket(self, account, pocket):
        self.assertTrue(pocket.amount, 'pocket %r has no amount' % pocket)
        self.assertTrue(pocket.label, 'pocket %r has no label' % pocket)

    def check_market_orders(self, account):
        if not isinstance(self.backend, CapBankWealth):
            return
        for market_order in self.backend.iter_market_orders(account):
            self.check_market_order(account, market_order)

    def check_market_order(self, account, market_order):
        self.assertTrue(market_order.label, 'Market order %r has no label' % market_order)
        self.assertFalse(
            empty(market_order.quantity) and empty(market_order.amount),
            'Market order %r has no quantity and no amount' % market_order
        )

    def check_recipients(self, account):
        if not isinstance(self.backend, CapTransfer):
            return
        for rcpt in self.backend.iter_transfer_recipients(account):
            self.check_recipient(account, rcpt)

    def check_recipient(self, account, rcpt):
        self.assertTrue(rcpt.id, 'recipient %r has no id' % rcpt)
        self.assertTrue(rcpt.label, 'recipient %r has no label' % rcpt)
        self.assertTrue(rcpt.category, 'recipient %r has no category' % rcpt)
        self.assertTrue(rcpt.enabled_at, 'recipient %r has no enabled_at' % rcpt)

    def check_emitters(self):
        if not isinstance(self.backend, CapTransfer):
            return
        for emitter in self.backend.iter_emitters():
            self.check_emitter(emitter)

    def check_emitter(self, emitter):
        self.assertTrue(emitter.id, 'emitter %r has no id' % emitter)
        # TODO



# Copyright(C) 2010-2011 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import logging
import time


__all__ = ['retry']


def retry(exceptions_to_check, exc_handler=None, tries=3, delay=2, backoff=2):
    """
    Retry decorator
    from https://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from https://wiki.python.org/moin/PythonDecoratorLibrary#Retry
    """
    def deco_retry(f):
        def f_retry(*args, **kwargs):
            mtries = kwargs.pop('_tries', tries)
            mdelay = kwargs.pop('_delay', delay)
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions_to_check as exc:
                    if exc_handler:
                        exc_handler(exc, **kwargs)
                    logging.debug('%s, Retrying in %d seconds...', exc, mdelay)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)
        return f_retry  # true decorator
    return deco_retry



# Copyright(C) 2014 Oleg Plakhotniuk
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from io import BytesIO, StringIO
from collections import namedtuple
import logging
import os
import subprocess
from tempfile import mkstemp


__all__ = ['decompress_pdf', 'get_pdf_rows']


def decompress_pdf(inpdf: bytes) -> bytes:
    """
    Takes PDF file contents as a string and returns decompressed version
    of the file contents, suitable for text parsing.

    External dependencies:
    MuPDF (https://www.mupdf.com).
    """

    inh, inname = mkstemp(suffix='.pdf')
    outh, outname = mkstemp(suffix='.pdf')
    os.write(inh, inpdf)
    os.close(inh)
    os.close(outh)

    subprocess.call(['mutool', 'clean', '-d', inname, outname])

    with open(outname, 'rb') as f:
        outpdf = f.read()
    os.remove(inname)
    os.remove(outname)
    return outpdf


Rect = namedtuple('Rect', ('x0', 'y0', 'x1', 'y1'))
TextRect = namedtuple('TextRect', ('x0', 'y0', 'x1', 'y1', 'text'))


def almost_eq(a, b):
    return abs(a - b) < 2


def lt_to_coords(obj, ltpage):
    # in a pdf, 'y' coords are bottom-to-top
    # in a pdf, coordinates are very often almost equal but not strictly equal

    x0 = (min(obj.x0, obj.x1))
    y0 = (min(ltpage.y1 - obj.y0, ltpage.y1 - obj.y1))
    x1 = (max(obj.x0, obj.x1))
    y1 = (max(ltpage.y1 - obj.y0, ltpage.y1 - obj.y1))

    x0 = round(x0)
    y0 = round(y0)
    x1 = round(x1)
    y1 = round(y1)

    # in a pdf, straight lines are actually rects, make them as thin as possible
    if almost_eq(x1, x0):
        x1 = x0
    if almost_eq(y1, y0):
        y1 = y0

    return Rect(x0, y0, x1, y1)


def lttext_to_multilines(obj, ltpage):
    # text lines within 'obj' are probably the same height
    x0 = (min(obj.x0, obj.x1))
    y0 = (min(ltpage.y1 - obj.y0, ltpage.y1 - obj.y1))
    x1 = (max(obj.x0, obj.x1))
    y1 = (max(ltpage.y1 - obj.y0, ltpage.y1 - obj.y1))

    lines = obj.get_text().rstrip('\n').split('\n')
    h = (y1 - y0) / len(lines)

    for n, line in enumerate(lines):
        yield TextRect((x0), (y0 + n * h), (x1), (y0 + n * h + h), line)


# fuzzy floats to smooth comparisons because lines are actually rects
# and seemingly-contiguous lines are actually not contiguous
class ApproxFloat(float):
    def __eq__(self, other):
        return almost_eq(self, other)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return self - other < 0 and self != other

    def __le__(self, other):
        return self - other <= 0 or self == other

    def __gt__(self, other):
        return not self <= other

    def __ge__(self, other):
        return not self < other


ANGLE_VERTICAL = 0
ANGLE_HORIZONTAL = 1
ANGLE_OTHER = 2


def angle(r):
    if r.x0 == r.x1:
        return ANGLE_VERTICAL
    elif r.y0 == r.y1:
        return ANGLE_HORIZONTAL
    return ANGLE_OTHER


class ApproxVecDict(dict):
    # since coords are never strictly equal, search coords around
    # store vectors and points

    def __getitem__(self, coords):
        x, y = coords
        for i in (0, -1, 1):
            for j in (0, -1, 1):
                try:
                    return super(ApproxVecDict, self).__getitem__((x+i, y+j))
                except KeyError:
                    pass
        raise KeyError()

    def get(self, k, v=None):
        try:
            return self[k]
        except KeyError:
            return v


class ApproxRectDict(dict):
    # like ApproxVecDict, but store rects
    def __getitem__(self, coords):
        x0, y0, x1, y1 = coords

        for i in (0, -1, 1):
            for j in (0, -1, 1):
                if x0 == x1:
                    for j2 in (0, -1, 1):
                        try:
                            return super(ApproxRectDict, self).__getitem__((x0+i, y0+j, x0+i, y1+j2))
                        except KeyError:
                            pass
                elif y0 == y1:
                    for i2 in (0, -1, 1):
                        try:
                            return super(ApproxRectDict, self).__getitem__((x0+i, y0+j, x1+i2, y0+j))
                        except KeyError:
                            pass
                else:
                    return super(ApproxRectDict, self).__getitem__((x0, y0, x1, y1))

        raise KeyError()


def uniq_lines(lines):
    new = ApproxRectDict()
    for line in lines:
        line = tuple(line)
        try:
            new[line]
        except KeyError:
            new[line] = None
    return [Rect(*k) for k in new.keys()]


def build_rows(lines):
    points = ApproxVecDict()

    # for each top-left point, build tuple with lines going down and lines going right
    for line in lines:
        a = angle(line)
        if a not in (ANGLE_HORIZONTAL, ANGLE_VERTICAL):
            continue

        coord = (line.x0, line.y0)
        plines = points.get(coord)
        if plines is None:
            plines = points[coord] = tuple([] for _ in range(2))

        plines[a].append(line)

    boxes = ApproxVecDict()
    for plines in points.values():
        if not (plines[ANGLE_HORIZONTAL] and plines[ANGLE_VERTICAL]):
            continue

        plines[ANGLE_HORIZONTAL].sort(key=lambda l: (l.y0, l.x1))
        plines[ANGLE_VERTICAL].sort(key=lambda l: (l.x0, l.y1))

        for hline in plines[ANGLE_HORIZONTAL]:
            try:
                vparallels = points[hline.x1, hline.y0][ANGLE_VERTICAL]
            except KeyError:
                continue
            if not vparallels:
                continue

            for vline in plines[ANGLE_VERTICAL]:
                try:
                    hparallels = points[vline.x0, vline.y1][ANGLE_HORIZONTAL]
                except KeyError:
                    continue
                if not hparallels:
                    continue

                hparallels = [hpar for hpar in hparallels if almost_eq(hpar.x1, hline.x1)]
                if not hparallels:
                    continue
                vparallels = [vpar for vpar in vparallels if almost_eq(vpar.y1, vline.y1)]
                if not vparallels:
                    continue

                assert len(hparallels) == 1 and len(vparallels) == 1
                assert almost_eq(hparallels[0].y0, vparallels[0].y1)
                assert almost_eq(vparallels[0].x0, hparallels[0].x1)

                box = Rect(hline.x0, hline.y0, hline.x1, vline.y1)
                boxes.setdefault((vline.y0, vline.y1), []).append(box)

    rows = list(boxes.values())
    new_rows = []
    for row in rows:
        row.sort(key=lambda box: box.x0)
        if row:
            row = [row[0]] + [c for n, c in enumerate(row[1:], 1) if row[n-1].x0 != c.x0]
        new_rows.append(row)

    rows = new_rows
    rows.sort(key=lambda row: row[0].y0)

    return rows


def find_in_table(rows, rect):
    for j, row in enumerate(rows):
        if ApproxFloat(row[0].y0) > rect.y1:
            break

        if not (ApproxFloat(row[0].y0) <= rect.y0 and ApproxFloat(row[0].y1) >= rect.y1):
            continue

        for i, box in enumerate(row):
            if ApproxFloat(box.x0) <= rect.x0 and ApproxFloat(box.x1) >= rect.x1:
                return i, j


def arrange_texts_in_rows(rows, trects):
    table = [[[] for _ in row] for row in rows]

    for trect in trects:
        pos = find_in_table(rows, trect)
        if not pos:
            continue
        table[pos[1]][pos[0]].append(trect.text)
    return table


LOGGER = logging.getLogger(__name__)
DEBUGFILES = logging.DEBUG - 1


def get_pdf_rows(data, miner_layout=True):
    """
    Takes PDF file content as string and yield table row data for each page.

    For each page in the PDF, the function yields a list of rows.
    Each row is a list of cells. Each cell is a list of strings present in the cell.
    Note that the rows may belong to different tables.

    There are no logic tables in PDF format, so this parses PDF drawing instructions
    and tries to find rectangles and arrange them in rows, then arrange text in
    the rectangles.

    External dependencies:
    PDFMiner (https://github.com/euske/pdfminer).
    """

    try:
        from pdfminer.pdfparser import PDFParser, PDFSyntaxError
    except ImportError:
        raise ImportError('Please install python3-pdfminer')

    try:
        from pdfminer.pdfdocument import PDFDocument
        from pdfminer.pdfpage import PDFPage
        newapi = True
    except ImportError:
        from pdfminer.pdfparser import PDFDocument
        newapi = False
    from pdfminer.converter import PDFPageAggregator
    from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
    from pdfminer.layout import LAParams, LTRect, LTTextBox, LTTextLine, LTLine, LTChar, LTCurve

    parser = PDFParser(BytesIO(data))
    try:
        if newapi:
            doc = PDFDocument(parser)
        else:
            doc = PDFDocument()
            parser.set_document(doc)
            doc.set_parser(parser)
    except PDFSyntaxError:
        return

    rsrcmgr = PDFResourceManager()
    if miner_layout:
        device = PDFPageAggregator(rsrcmgr, laparams=LAParams())
    else:
        device = PDFPageAggregator(rsrcmgr)

    interpreter = PDFPageInterpreter(rsrcmgr, device)
    if newapi:
        pages = PDFPage.get_pages(BytesIO(data), check_extractable=True)
    else:
        doc.initialize()
        pages = doc.get_pages()

    if LOGGER.isEnabledFor(DEBUGFILES):
        import tempfile
        import PIL.Image as Image
        import PIL.ImageDraw as ImageDraw
        import random

        path = tempfile.mkdtemp(prefix='pdf')

    for npage, page in enumerate(pages):
        LOGGER.debug('processing page %s', npage)
        interpreter.process_page(page)
        page_layout = device.get_result()

        texts = sum([list(lttext_to_multilines(obj, page_layout)) for obj in page_layout._objs if isinstance(obj, (LTTextBox, LTTextLine, LTChar))], [])
        LOGGER.debug('found %d text objects', len(texts))
        if LOGGER.isEnabledFor(DEBUGFILES):
            img = Image.new('RGB', (int(page.mediabox[2]), int(page.mediabox[3])), (255, 255, 255))
            draw = ImageDraw.Draw(img)
            for t in texts:
                color = (random.randint(127, 255), random.randint(127, 255), random.randint(127, 255))
                draw.rectangle((t.x0, t.y0, t.x1, t.y1), outline=color)
                draw.text((t.x0, t.y0), t.text.encode('utf-8'), color)
            fpath = '%s/1text-%03d.png' % (path, npage)
            img.save(fpath)
            LOGGER.log(DEBUGFILES, 'saved %r', fpath)

        if not miner_layout:
            texts.sort(key=lambda t: (t.y0, t.x0))

        # TODO filter ltcurves that are not lines?
        # TODO convert rects to 4 lines?
        lines = [lt_to_coords(obj, page_layout) for obj in page_layout._objs if isinstance(obj, (LTRect, LTLine, LTCurve))]
        LOGGER.debug('found %d lines', len(lines))
        if LOGGER.isEnabledFor(DEBUGFILES):
            img = Image.new('RGB', (int(page.mediabox[2]), int(page.mediabox[3])), (255, 255, 255))
            draw = ImageDraw.Draw(img)
            for l in lines:
                color = (random.randint(127, 255), random.randint(127, 255), random.randint(127, 255))
                draw.rectangle((l.x0, l.y0, l.x1, l.y1), outline=color)
            fpath = '%s/2lines-%03d.png' % (path, npage)
            img.save(fpath)
            LOGGER.log(DEBUGFILES, 'saved %r', fpath)

        lines = list(uniq_lines(lines))
        LOGGER.debug('found %d unique lines', len(lines))

        rows = build_rows(lines)
        LOGGER.debug('built %d rows (%d boxes)', len(rows), sum(len(row) for row in rows))
        if LOGGER.isEnabledFor(DEBUGFILES):
            img = Image.new('RGB', (int(page.mediabox[2]), int(page.mediabox[3])), (255, 255, 255))
            draw = ImageDraw.Draw(img)
            for r in rows:
                for b in r:
                    color = (random.randint(127, 255), random.randint(127, 255), random.randint(127, 255))
                    draw.rectangle((b.x0 + 1, b.y0 + 1, b.x1 - 1, b.y1 - 1), outline=color)
            fpath = '%s/3rows-%03d.png' % (path, npage)
            img.save(fpath)
            LOGGER.log(DEBUGFILES, 'saved %r', fpath)

        textrows = arrange_texts_in_rows(rows, texts)
        LOGGER.debug('assigned %d strings', sum(sum(len(c) for c in r) for r in textrows))
        if LOGGER.isEnabledFor(DEBUGFILES):
            img = Image.new('RGB', (int(page.mediabox[2]), int(page.mediabox[3])), (255, 255, 255))
            draw = ImageDraw.Draw(img)
            for row, trow in zip(rows, textrows):
                for b, tlines in zip(row, trow):
                    color = (random.randint(127, 255), random.randint(127, 255), random.randint(127, 255))
                    draw.rectangle((b.x0 + 1, b.y0 + 1, b.x1 - 1, b.y1 - 1), outline=color)
                    draw.text((b.x0 + 1, b.y0 + 1), '\n'.join(tlines).encode('utf-8'), color)
            fpath = '%s/4cells-%03d.png' % (path, npage)
            img.save(fpath)
            LOGGER.log(DEBUGFILES, 'saved %r', fpath)

        yield textrows
    device.close()

# Export part #

def html_to_pdf(browser, url=None, data=None, extra_options=None):
    """
    Convert html to PDF.

    :param browser: browser instance
    :param url: link to the html ressource
    :param data: HTML content
    :return: the document converted in PDF
    :rtype: bytes
    """
    try:
        import pdfkit # https://pypi.python.org/pypi/pdfkit
    except ImportError:
        raise ImportError('Please install python3-pdfkit')

    assert (url or data) and not (url and data), 'Please give only url or data parameter'

    callback = pdfkit.from_url if url else pdfkit.from_string
    options = {}

    try:
        cookies = browser.session.cookies
    except AttributeError:
        pass
    else:
        options.update({
            'cookie': [(cookie, value) for cookie, value in cookies.items() if value], # cookies of browser
        })

    if extra_options:
        options.update(extra_options)

    return callback(url or data, False, options=options)


class BlinkPdfError(Exception):
    pass


def blinkpdf(browser, url, extra_options=None, filter_cookie=None, start_xvfb=True, timeout=120):
    # - xvfb is required for blinkpdf 1.0, but not for 1.1
    # - xvfb is not necessary for QtWebEngine 5.14, but it is for 5.11, which is the version
    #   available on the ppa for debian/buster stable

    xvfb_exists = False
    blinkpdf_exists = False
    paths = os.getenv('PATH', os.defpath).split(os.pathsep)
    for path in paths:
        fpath = os.path.join(path, 'xvfb-run')
        if os.path.exists(fpath) and os.access(fpath, os.X_OK):
            xvfb_exists = True
        fpath = os.path.join(path, 'blinkpdf')
        if os.path.exists(fpath) and os.access(fpath, os.X_OK):
            blinkpdf_exists = True

    if (not xvfb_exists and start_xvfb) or not blinkpdf_exists:
        raise NotImplementedError()

    args = []
    for c in browser.session.cookies:
        if c.value:
            if not filter_cookie or filter_cookie(c):
                args.append('--cookie')
                args.append('%s=%s' % (c.name, c.value))

    for key, value in browser.session.headers.items():
        args.append('--header')
        args.append('%s=%s' % (key, value))

    if extra_options and 'run-script' in extra_options:
        args.append('--run-script')
        args.append(extra_options['run-script'][0])

    args.append(url)
    args.append('-')  # - : don't write it on disk, simply return value

    if start_xvfb:
        # put a very small resolution to reduce used memory, because we don't really need it, it doesn't influence pdf size
        # -screen 0 width*height*bit depth
        prepend = ['xvfb-run', '-a', '-s', '-screen 0 2x2x8', 'blinkpdf']
    else:
        prepend = ['blinkpdf']

    cmd = list(prepend) + list(args)

    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        # Will raise a TimeoutExpired after timeout seconds
        stdout, stderr = proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        # A timeout doesn't kill the child process
        proc.kill()
        # Log the error output after the end of the process. 20 seconds should
        # be enough for the process to terminate cleanly
        _, stderr = proc.communicate(timeout=20)
        LOGGER.error('The blinkpdf process took too long to complete. Error output: %s', stderr.decode('utf-8'))
        raise

    if proc.returncode != 0:
        raise BlinkPdfError('command returned non-zero exit status 1: %s' % stderr.decode('utf-8'))
    return stdout


# extract all text from PDF
def extract_text(data):
    try:
        try:
            from pdfminer.pdfdocument import PDFDocument
            from pdfminer.pdfpage import PDFPage
            newapi = True
        except ImportError:
            from pdfminer.pdfparser import PDFDocument
            newapi = False
        from pdfminer.pdfparser import PDFParser, PDFSyntaxError
        from pdfminer.converter import TextConverter
        from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
    except ImportError:
        raise ImportError('Please install python3-pdfminer to parse PDF')
    else:
        parser = PDFParser(BytesIO(data))
        try:
            if newapi:
                doc = PDFDocument(parser)
            else:
                doc = PDFDocument()
                parser.set_document(doc)
                doc.set_parser(parser)
        except PDFSyntaxError:
            return

        rsrcmgr = PDFResourceManager()
        out = StringIO()
        device = TextConverter(rsrcmgr, out)
        interpreter = PDFPageInterpreter(rsrcmgr, device)
        if newapi:
            pages = PDFPage.create_pages(doc)
        else:
            doc.initialize()
            pages = doc.get_pages()
        for page in pages:
            interpreter.process_page(page)

        return out.getvalue()



# Copyright(C) 2023 Powens
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import mimetypes
from os.path import splitext
from typing import Any

from woob.browser.filters.standard import CleanText, FormatError
from woob.capabilities.base import empty
from woob.tools.misc import NO_DEFAULT

from .base import debug

__all__ = ['MimeType', 'FileExtension']


class MimeType(CleanText):
    """
    A filter to determine the MIME Type (Multipurpose Internet Mail Extensions)
    of a file based on a given string, which can be a file path or a file name with an extension.

    :param default: The default MIME type to be returned when the file type is not recognized.
    :type default: Any, optional
    """
    @debug()
    def filter(self, txt: str) -> Any:
        """
        Get the MIME type from a file name or path.

        :param txt: The file name or path for which to determine the MIME type.
        :type txt: str
        :raises FormatError: If the MIME type is not recognized.

        >>> MimeType().filter('foo.pdf')
        'application/pdf'
        >>> MimeType().filter('path/foo/invoices.tar.gz')
        'application/x-tar'
        >>> MimeType(default='NAN').filter('foo.no')
        'NAN'
        """
        txt = super().filter(txt)

        if empty(txt):
            return self.default_or_raise(FormatError(f'Unable to parse {txt}'))
        # The 'mimetypes.guess_type()' function requires a valid
        # file name with an extension (file_name.extension)
        # to determine the MIME type. It may not handle inputs without a dot ('pdf'),
        # or with a dot but no name ('.pdf').
        if txt.startswith('.'):
            # .pdf
            txt = f'dummy_filename{txt}'
        if '.' not in txt:
            # pdf
            txt = f'dummy_filename.{txt}'

        mime_type, _ = mimetypes.guess_type(txt)
        if not mime_type:
            return self.default_or_raise(
                FormatError(f'MIME type not recognized for file: {txt}')
            )
        return mime_type


class FileExtension(CleanText):
    """
    A filter to extract the file extension from a given string representing
    a file name or a file path.

    :param default: The default extension to be returned when the file extension is not recognized.
    :type default: Any, optional
    :param validate_mime: Flag to indicate whether to validate the MIME type of the returned extension.
    :type validate_mime: bool, optional
    """

    def __init__(self, selector=None, validate_mime=False, default=NO_DEFAULT):
        super().__init__(selector, default=default)
        self.validate_mime = validate_mime

    @debug()
    def filter(self, txt: str) -> Any:
        """
        Get the file extension from a file name or path.

        :param txt: The file name or path for which to extract the file extension.
        :type txt: str
        :raises FormatError: If the file extension is not recognized.

        >>> FileExtension().filter('file.docx')
        'docx'
        >>> FileExtension().filter('path/to/file.tar.gz')
        'tar.gz'
        >>> FileExtension(default='NAN').filter('file_without_extension')
        'NAN'
        >>> FileExtension().filter('/home/user/Documents/report.pdf')
        'pdf'
        >>> FileExtension(default='UNKNOWN').filter('spreadsheet')
        'UNKNOWN'
        >>> FileExtension(default='UNKNOWN', validate_mime=True).filter('path/to/file.dfs')
        'UNKNOWN'
        >>> FileExtension(default='UNKNOWN', validate_mime=True).filter('file.jpg')
        'jpg'
        """
        txt = super().filter(txt)
        if empty(txt):
            return self.default_or_raise(FormatError(f'Unable to parse {txt}'))

        if len(txt.split('.')) > 2:
            extension = '.'.join(txt.split('.')[-2:])
        else:
            _, extension = splitext(txt)
            if not extension:
                return self.default_or_raise(FormatError(f'Extension not recognized: {txt}'))
            extension = extension.strip('.')

        if self.validate_mime:
            try:
                MimeType().filter(extension)
            except FormatError:
                return self.default_or_raise(FormatError(
                    f'MIME type not recognized for the extension {extension}')
                )
        return extension



# Copyright(C) 2013 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.lyrics import CapLyrics
from woob.capabilities.base import empty
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter


__all__ = ['AppLyrics', 'LyricsGetFormatter', 'LyricsListFormatter']


class LyricsGetFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'artist', 'content')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.title, self.NC)
        result += 'ID: %s\n' % obj.fullid
        result += 'Title: %s\n' % obj.title
        result += 'Artist: %s\n' % obj.artist
        result += '\n%sContent%s\n' % (self.BOLD, self.NC)
        result += '%s' % obj.content
        return result


class LyricsListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'artist')

    def get_title(self, obj):
        return obj.title

    def get_description(self, obj):
        artist = ''
        if not empty(obj.artist):
            artist = obj.artist
        return '%s' % artist


class AppLyrics(ReplApplication):
    APPNAME = 'lyrics'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2013-YEAR Julien Veyssier'
    DESCRIPTION = "Console application allowing to search for song lyrics on various websites."
    SHORT_DESCRIPTION = "search and display song lyrics"
    CAPS = CapLyrics
    EXTRA_FORMATTERS = {'lyrics_list': LyricsListFormatter,
                        'lyrics_get': LyricsGetFormatter,
                        }
    COMMANDS_FORMATTERS = {'search':    'lyrics_list',
                           'get':      'lyrics_get',
                           }
    SEARCH_CRITERIAS = ['artist', 'song']

    def complete_get(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_get(self, id):
        """
        get ID

        Display lyrics of the song.
        """

        songlyrics = self.get_object(id, 'get_lyrics')
        if not songlyrics:
            print('Song lyrics not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(songlyrics)

    def complete_search(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self.SEARCH_CRITERIAS

    @defaultcount(10)
    def do_search(self, line):
        """
        search [artist | song] [PATTERN]

        Search lyrics by artist name or by song title.
        """
        criteria, pattern = self.parse_command_args(line, 2, 2)
        self.change_path(['search'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for songlyrics in self.do('iter_lyrics', criteria, pattern):
            self.cached_format(songlyrics)



# Copyright(C) 2014 Simon Murail
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

try:
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    ThreadPoolExecutor = None

from http import cookiejar

from requests import Session
from requests.adapters import DEFAULT_POOLSIZE
from requests.cookies import RequestsCookieJar, cookiejar_from_dict
from requests.models import PreparedRequest
from requests.sessions import merge_setting
from requests.structures import CaseInsensitiveDict
from requests.utils import get_netrc_auth

from .adapters import HTTPAdapter


def merge_hooks(request_hooks, session_hooks):
    """
    Properly merges both requests and session hooks.

    This is necessary because when request_hooks == {'response': []}, the
    merge breaks Session hooks entirely.

    Backport from request so we can use it in wheezy
    """
    if session_hooks is None or session_hooks.get('response') == []:
        return request_hooks

    if request_hooks is None or request_hooks.get('response') == []:
        return session_hooks

    ret = {}
    for (k, v) in request_hooks.items():
        if v is not None:
            ret[k] = set(v).union(session_hooks.get(k, []))

    return ret


class WoobSession(Session):
    def prepare_request(self, request):
        """Constructs a :class:`PreparedRequest <PreparedRequest>` for
        transmission and returns it. The :class:`PreparedRequest` has settings
        merged from the :class:`Request <Request>` instance and those of the
        :class:`Session`.

        :param request: :class:`Request` instance to prepare with this
                        session's settings.
        """
        cookies = request.cookies or {}

        # Bootstrap CookieJar.
        if not isinstance(cookies, cookiejar.CookieJar):
            cookies = cookiejar_from_dict(cookies)

        # Merge with session cookies
        merged_cookies = RequestsCookieJar()
        merged_cookies.update(self.cookies)
        merged_cookies.update(cookies)

        # Set environment's basic authentication if not explicitly set.
        auth = request.auth
        if self.trust_env and not auth and not self.auth:
            auth = get_netrc_auth(request.url)

        p = PreparedRequest()
        p.prepare(
            method=request.method.upper(),
            url=request.url,
            files=request.files,
            data=request.data,
            json=request.json,
            headers=merge_setting(request.headers, self.headers, dict_class=CaseInsensitiveDict),
            params=merge_setting(request.params, self.params),
            auth=merge_setting(auth, self.auth),
            cookies=merged_cookies,
            hooks=merge_hooks(request.hooks, self.hooks),
        )
        return p


WeboobSession = WoobSession


class FuturesSession(WoobSession):
    def __init__(self, executor=None, max_workers=2, max_retries=2, adapter_class=HTTPAdapter, *args, **kwargs):
        """Creates a FuturesSession

        Notes
        ~~~~~

        * ProcessPoolExecutor is not supported b/c Response objects are
          not picklable.

        * If you provide both `executor` and `max_workers`, the latter is
          ignored and provided executor is used as is.
        """
        super(FuturesSession, self).__init__(*args, **kwargs)
        if executor is None and ThreadPoolExecutor is not None:
            executor = ThreadPoolExecutor(max_workers=max_workers)
            # set connection pool size equal to max_workers if needed
            if max_workers > DEFAULT_POOLSIZE:
                adapter_kwargs = dict(pool_connections=max_workers,
                                      pool_maxsize=max_workers,
                                      max_retries=max_retries)
                self.mount('https://', adapter_class(**adapter_kwargs))
                self.mount('http://', adapter_class(**adapter_kwargs))

        self.executor = executor

    def send(self, *args, **kwargs):
        """Maintains the existing api for :meth:`Session.send`

        Used by :meth:`request` and thus all of the higher level methods

        If the `is_async` param is True, the request is processed in a
        thread. Otherwise, the request is processed as usual, in a blocking way.

        In all cases, it will call the `callback` parameter and return its
        result when the request has been processed.
        """
        if 'async' in kwargs:
            import warnings
            warnings.warn('Please use is_async instead of async.', DeprecationWarning)
            kwargs['is_async'] = kwargs['async']
            del kwargs['async']

        sup = super(FuturesSession, self).send

        callback = kwargs.pop('callback', lambda future, response: response)
        is_async = kwargs.pop('is_async', False)

        def func(*args, **kwargs):
            resp = sup(*args, **kwargs)
            return callback(self, resp)

        if is_async:
            if not self.executor:
                raise ImportError('Please install python3-concurrent.futures')
            return self.executor.submit(func, *args, **kwargs)

        return func(*args, **kwargs)

    def close(self):
        super(FuturesSession, self).close()
        if self.executor:
            self.executor.shutdown()



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import re
import datetime
from collections import OrderedDict
from typing import TypeVar

from .misc import to_unicode


__all__ = ['ValuesDict', 'Value', 'ValueBackendPassword', 'ValueInt', 'ValueFloat', 'ValueBool']

ValuesDictType = TypeVar('ValuesDictType', bound='ValuesDict')


class ValuesDict(OrderedDict):
    """Ordered dictionary which can take values in constructor.

    >>> ValuesDict(Value('a', label='Test'), ValueInt('b', label='Test2'))
    """

    def __init__(self, *values):
        super(ValuesDict, self).__init__()
        for v in values:
            self[v.id] = v

    def with_values(self: ValuesDictType, *values: Value) -> ValuesDictType:
        """Get a copy of the object, with new values.

        :param values: The values to set.
        :return: The new values dictionary.
        """
        existing_values = {key: value for key, value in self.items()}
        existing_values.update({value.id: value for value in values})
        return self.__class__(*existing_values.values())

    def with_values_from(self: ValuesDictType, other: ValuesDict) -> ValuesDictType:
        """Get a copy of the object, with overrides from another values dictionary.

        Values from the other dictionary will override values from the
        current dictionary.

        :param other: the other dictionary to take values from.
        :return: The new values dictionary.
        """
        return self.with_values(*other.values())

    def without_values(self: ValuesDictType, *value_names: str) -> ValuesDictType:
        """Get a copy of the object, without values with the given names.

        This method will ignore value names that aren't present in the
        original dictionary.

        :param value_names: The name of the values to remove.
        :return: The new values dictionary.
        """
        existing_values = {key: value for key, value in self.items()}
        for value_name in value_names:
            existing_values.pop(value_name, None)

        return self.__class__(*existing_values.values())


class Value:
    """
    Value.

    :param label: human readable description of a value
    :type label: str
    :param required: if ``True``, the backend can't load if the key isn't found in its configuration
    :type required: bool
    :param default: an optional default value, used when the key is not in config. If there is no default value and the key
                    is not found in configuration, the **required** parameter is implicitly set
    :param masked: if ``True``, the value is masked. It is useful for applications to know if this key is a password
    :type masked: bool
    :param regexp: if specified, on load the specified value is checked against this regexp, and an error is raised if it doesn't match
    :type regexp: str
    :param choices: if this parameter is set, the value must be in the list
    :type choices: (list,dict)
    :param aliases: mapping of old choices values that should be accepted but not presented
    :type aliases: dict
    :param tiny: the value of choices can be entered by an user (as they are small)
    :type tiny: bool
    :param transient: this value is not persistent (asked only if needed)
    :type transient: bool
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            self.id = args[0]
        else:
            self.id = ''
        self.label = kwargs.get('label', kwargs.get('description', None))
        self.description = kwargs.get('description', kwargs.get('label', None))
        self.default = kwargs.get('default', None)
        if isinstance(self.default, str):
            self.default = to_unicode(self.default)
        self.regexp = self.get_normalized_regexp(kwargs.get('regexp', None))
        self.choices = kwargs.get('choices', None)
        self.aliases = kwargs.get('aliases')
        if isinstance(self.choices, (list, tuple)):
            self.choices = OrderedDict(((v, v) for v in self.choices))
        self.tiny = kwargs.get('tiny', None)
        self.transient = kwargs.get('transient', None)
        self.masked = kwargs.get('masked', False)
        self.required = kwargs.get('required', self.default is None)
        self._value = kwargs.get('value', None)

    @staticmethod
    def get_normalized_regexp(regexp):
        """ Return normalized regexp adding missing anchors """

        if not regexp:
            return regexp
        if not regexp.startswith('^'):
            regexp = '^' + regexp
        if not regexp.endswith('$'):
            regexp += '$'
        return regexp

    def show_value(self, v):
        if self.masked:
            return ''
        else:
            return v

    def check_valid(self, v):
        """
        Check if the given value is valid.

        :raises: ValueError
        """
        if self.required and v is None:
            raise ValueError('Value is required and thus must be set')
        if v == self.default:
            return
        if v == '' and self.default != '' and (self.choices is None or v not in self.choices):
            raise ValueError('Value can\'t be empty')
        if self.regexp is not None and not re.match(self.regexp, str(v) if v is not None else ''):
            raise ValueError('Value does not match regexp "%s"' % self.regexp)
        if self.choices is not None and v not in self.choices:
            if not self.aliases or v not in self.aliases:
                raise ValueError(
                    'Value is not in list: %s' % (
                        ', '.join(str(s) for s in self.choices)
                    )
                )

    def load(self, domain, v, requests):
        """
        Load value.

        :param domain: what is the domain of this value
        :type domain: str
        :param v: value to load
        :param requests: list of woob requests
        :type requests: woob.core.requests.Requests
        """
        return self.set(v)

    def set(self, v):
        """
        Set a value.
        """
        self.check_valid(v)
        if self.aliases and v in self.aliases:
            v = self.aliases[v]
        self._value = v

    def dump(self):
        """
        Dump value to be stored.
        """
        return self.get()

    def get(self):
        """
        Get the value.
        """
        return self._value


class ValueTransient(Value):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('transient', True)
        kwargs.setdefault('default', None)
        kwargs.setdefault('required', False)
        super(ValueTransient, self).__init__(*args, **kwargs)

    def dump(self):
        return ''


class ValueBackendPassword(Value):
    _domain = None
    _requests = None
    _stored = True

    def __init__(self, *args, **kwargs):
        kwargs['masked'] = kwargs.pop('masked', True)
        self.noprompt = kwargs.pop('noprompt', False)
        super(ValueBackendPassword, self).__init__(*args, **kwargs)
        self.default = kwargs.get('default', '')

    def load(self, domain, password, requests):
        self.check_valid(password)
        self._domain = domain
        self._value = to_unicode(password)
        self._requests = requests

    def check_valid(self, passwd):
        if passwd == '':
            # always allow empty passwords
            return True
        return super(ValueBackendPassword, self).check_valid(passwd)

    def set(self, passwd):
        self.check_valid(passwd)
        if passwd is None:
            # no change
            return
        self._value = ''
        if passwd == '':
            return
        if self._domain is None:
            self._value = to_unicode(passwd)
            return

        self._value = to_unicode(passwd)

    def dump(self):
        if self._stored:
            return self._value
        else:
            return ''

    def get(self):
        if self._value != '' or self._domain is None:
            return self._value

        passwd = None

        if passwd is not None:
            # Password has been read in the keyring.
            return to_unicode(passwd)

        # Prompt user to enter password by hand.
        if not self.noprompt and self._requests:
            self._value = self._requests.request('login', self._domain, self)
            if self._value is None:
                self._value = ''
            else:
                self._value = to_unicode(self._value)
                self._stored = False
        return self._value


class ValueInt(Value):
    def __init__(self, *args, **kwargs):
        kwargs['regexp'] = r'^\d+$'
        super(ValueInt, self).__init__(*args, **kwargs)
        self.default = kwargs.get('default', 0)

    def get(self):
        return int(self._value)


class ValueFloat(Value):
    def __init__(self, *args, **kwargs):
        kwargs['regexp'] = r'^[\d\.]+$'
        super(ValueFloat, self).__init__(*args, **kwargs)
        self.default = kwargs.get('default', 0.0)

    def check_valid(self, v):
        try:
            float(v)
        except ValueError:
            raise ValueError('Value is not a float value')

    def get(self):
        return float(self._value)


class ValueBool(Value):
    def __init__(self, *args, **kwargs):
        kwargs['choices'] = {'y': 'True', 'n': 'False'}
        super(ValueBool, self).__init__(*args, **kwargs)
        self.default = kwargs.get('default', False)

    def check_valid(self, v):
        if not isinstance(v, bool) and \
            str(v).lower() not in {
                'y', 'yes', '1', 'true',  'on',
                'n', 'no',  '0', 'false', 'off',
            }:

            raise ValueError('Value is not a boolean (y/n)')

    def get(self):
        return (isinstance(self._value, bool) and self._value) or \
                str(self._value).lower() in {'y', 'yes', '1', 'true', 'on'}


class ValueDate(Value):
    DEFAULT_FORMAT = '%Y-%m-%d'

    def __init__(self, *args, **kwargs):
        formats = tuple(kwargs.pop('formats', ()))
        super(ValueDate, self).__init__(*args, **kwargs)

        if formats:
            self.preferred_format = formats[0]
        else:
            self.preferred_format = self.DEFAULT_FORMAT
        self.accepted_formats = (self.DEFAULT_FORMAT,) + formats

    def _parse(self, v):
        for format in self.accepted_formats:
            try:
                dateval = datetime.datetime.strptime(v, format).date()
            except ValueError:
                continue
            return dateval

        raise ValueError('Value does not match format in %s' % self.accepted_formats)

    def check_valid(self, v):
        if self.required and not v:
            raise ValueError('Value is required and thus must be set')

    def load(self, domain, v, requests):
        self.check_valid(v)
        if not v:
            self._value = None
            return
        if isinstance(v, str):
            v = self._parse(v)
        if isinstance(v, datetime.date):
            self._value = v
        else:
            raise ValueError('Value is not of the proper type')

    def dump(self):
        if self._value:
            return self._value.strftime(self.DEFAULT_FORMAT)

    def set(self, v):
        self.load(None, v, None)

    def get_as_string(self):
        if not self._value:
            return self._value

        return self._value.strftime(self.preferred_format)
